from __future__ import annotations

import math
from collections import namedtuple

from scitbx import matrix

from dials.algorithms.refinement.parameterisation.prediction_parameters import (
    SparseGradientVectorMixin,
    XYPhiPredictionParameterisation,
)
from dials.array_family import flex
from dials_refinement_helpers_ext import (
    build_reconstitute_derivatives_mat3,
    build_reconstitute_derivatives_vec3,
    intersection_i_seqs_unsorted,
)


class SparseFlex:
    """A wrapper for flex arrays that allows sparse storage by recording the
    values as a dense array, the length of the sparse array and the indices
    of the values into the sparse array. This is designed as a simple means
    to achieve sparse storage of flex mat3 and vec3 arrays and allows some
    operations to be performed with flex arrays and other SparseFlex arrays.

    The operations that can be performed are purposely limited. For example,
    addition and subtraction are only performed between two SparseFlex arrays,
    where it is assumed (not tested) that these have the same pattern of
    structural zeroes."""

    def __init__(self, dimension, elements, indices):

        if len(elements) != len(indices):
            raise ValueError(
                "The arrays of elements and indices must be of equal length"
            )
        self._size = dimension
        self._data = elements
        self._indices = indices

    def select(self, indices):

        try:
            indices = indices.iselection()
        except AttributeError:
            pass

        if self._data is None:
            return self

        # New object must have the dimension of the selection
        dimension = len(indices)

        # Calculate the intersection of these indices
        index_a, index_b = intersection_i_seqs_unsorted(self._indices, indices)

        # The first set of indices select the data, while the second set
        # provide their new indices
        elements = self._data.select(index_a)
        return SparseFlex(dimension, elements, index_b)

    @property
    def size(self):
        return self._size

    @property
    def non_zeroes(self):
        return len(self._data)

    def as_dense_vector(self):
        v = self._data.deep_copy()
        v *= 0.0
        v.resize(self._size)
        v.set_selected(self._indices, self._data)
        return v

    @property
    def data_and_indices(self):
        return (self._data, self._indices)

    def _extract_explicit_data(self, other):
        """Return the flex array of explicit data elements if other is a flex
        array or a SparseFlex"""

        # Take only the explicit data if other is a SparseFlex
        if isinstance(other, SparseFlex):
            return other._data

        # Otherwise select only explicit elements from a flex array
        try:
            other = other.select(self._indices)
        except AttributeError:
            pass

        return other

    def __mul__(self, other):

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, self._data * other, self._indices)

    def __rmul__(self, other):

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, other * self._data, self._indices)

    def __truediv__(self, other):

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, self._data / other, self._indices)

    def __add__(self, other):

        if not isinstance(other, SparseFlex):
            raise TypeError("Addition is only defined between two SparseFlex arrays")

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, self._data + other, self._indices)

    def __sub__(self, other):

        if not isinstance(other, SparseFlex):
            raise TypeError("Subtraction is only defined between two SparseFlex arrays")

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, self._data - other, self._indices)

    def dot(self, other):

        other = self._extract_explicit_data(other)

        return SparseFlex(self._size, self._data.dot(other), self._indices)

    def rotate_around_origin(self, direction, angle):

        angle = self._extract_explicit_data(angle)
        direction = self._extract_explicit_data(direction)
        return SparseFlex(
            self._size, self._data.rotate_around_origin(direction, angle), self._indices
        )

    def parts(self):

        x, y, z = self._data.parts()
        return (
            SparseFlex(self._size, x, self._indices),
            SparseFlex(self._size, y, self._indices),
            SparseFlex(self._size, z, self._indices),
        )


class StateDerivativeCache:
    """Keep derivatives of the model states in a memory-efficient format
    by storing each derivative once alongside the indices of reflections affected
    by that derivative"""

    def __init__(self, parameterisations=None):

        if parameterisations is None:
            parameterisations = []
        self._cache = dict.fromkeys(parameterisations)

        self._Pair = namedtuple("Pair", ["derivative", "iselection"])

        # set up lists with the right number of elements
        self.clear()

        self._nref = 0

    def build_gradients(self, parameterisation, isel=None, imatch=None):
        """Return an object mimicking a list of flex arrays containing state
        gradients wrt each parameter of the parameterisation. In fact this is a
        generator so that iterating over the elements of the list will return
        control here so that the gradient array for a single parameter can be
        reconstructed on the fly"""

        # Get the data from the cache
        entry = self._cache[parameterisation]

        # Figure out the right flex array type from entries in the cache
        shape = None
        for e in entry:
            if e:
                shape = e[0].derivative.n
                break
        if shape is None:
            raise TypeError("No model state derivatives found")
        if shape == (3, 1):
            build = build_reconstitute_derivatives_vec3
        elif shape == (3, 3):
            build = build_reconstitute_derivatives_mat3
        else:
            raise TypeError("Unrecognised model state derivative type")

        # Loop over the data for each parameter
        for p_data in entry:

            # Reconstitute full array from the cache and pack into a SparseFlex
            total_nelem = sum(pair.iselection.size() for pair in p_data)
            recon = build(total_nelem)
            for pair in p_data:
                recon.add_data(pair.derivative, pair.iselection)
            ds_dp = SparseFlex(self._nref, recon.get_data(), recon.get_indices())

            # First select only elements relevant to the current gradient calculation
            # block (i.e. if nproc > 1 or gradient_calculation_blocksize was set)
            if imatch is not None:
                ds_dp = ds_dp.select(imatch)

            # Now select only those reflections from the full list that are affected
            # by this parameterisation
            if isel is not None:
                ds_dp = ds_dp.select(isel)

            yield ds_dp

    def clear(self):
        """Clear all cached values"""

        for p in self._cache:
            self._cache[p] = [[] for i in range(p.num_free())]

    def append(self, parameterisation, iparam, derivative, iselection):
        """For a particular parameterisation and parameter number of the free
        parameters of that parameterisation, append a state derivative and the
        iselection of reflections it affects to the cache"""

        l1 = self._cache[parameterisation]
        l2 = l1[iparam]
        l2.append(self._Pair(derivative, iselection))

    @property
    def nref(self):
        """Get the length of the reflection list to which indices in the iselections
        refer"""

        return self._nref

    @nref.setter
    def nref(self, value):
        """Set the length of the reflection list to which indices in the iselections
        refer"""

        self._nref = value


class ScanVaryingPredictionParameterisation(XYPhiPredictionParameterisation):
    """An extension of the rotation scans version of the
    PredictionParameterisation class that supports model parameterisations that
    vary smoothly with the observed image number"""

    def __init__(
        self,
        experiments,
        detector_parameterisations=None,
        beam_parameterisations=None,
        xl_orientation_parameterisations=None,
        xl_unit_cell_parameterisations=None,
        goniometer_parameterisations=None,
    ):

        if detector_parameterisations is None:
            detector_parameterisations = []
        if beam_parameterisations is None:
            beam_parameterisations = []
        if xl_orientation_parameterisations is None:
            xl_orientation_parameterisations = []
        if xl_unit_cell_parameterisations is None:
            xl_unit_cell_parameterisations = []
        if goniometer_parameterisations is None:
            goniometer_parameterisations = []

        # determine once here which types of parameterisations are scan-varying
        self._varying_detectors = any(
            hasattr(p, "num_sets") for p in detector_parameterisations
        )
        self._varying_beams = any(
            hasattr(p, "num_sets") for p in beam_parameterisations
        )
        self._varying_xl_orientations = any(
            hasattr(p, "num_sets") for p in xl_orientation_parameterisations
        )
        self._varying_xl_unit_cells = any(
            hasattr(p, "num_sets") for p in xl_unit_cell_parameterisations
        )
        self._varying_goniometers = any(
            hasattr(p, "num_sets") for p in goniometer_parameterisations
        )

        to_cache = []
        if self._varying_detectors:
            to_cache.extend(detector_parameterisations)
        if self._varying_beams:
            to_cache.extend(beam_parameterisations)
        if self._varying_xl_orientations:
            to_cache.extend(xl_orientation_parameterisations)
        if self._varying_xl_unit_cells:
            to_cache.extend(xl_unit_cell_parameterisations)
        if self._varying_goniometers:
            to_cache.extend(goniometer_parameterisations)
        self._derivative_cache = StateDerivativeCache(to_cache)

        # set up base class
        super().__init__(
            experiments,
            detector_parameterisations=detector_parameterisations,
            beam_parameterisations=beam_parameterisations,
            xl_orientation_parameterisations=xl_orientation_parameterisations,
            xl_unit_cell_parameterisations=xl_unit_cell_parameterisations,
            goniometer_parameterisations=goniometer_parameterisations,
        )

        # Avoid calculation in calculate_model_state_uncertainties unless this
        # is set to True
        self.set_scan_varying_errors = False

    def _get_xl_orientation_parameterisation(self, experiment_id):
        """Return the crystal orientation parameterisation for the requested
        experiment number (or None if the crystal orientation in that experiment
        is not parameterised)"""

        param_set = self._exp_to_param[experiment_id]
        xl_op = None
        if param_set.xl_ori_param is not None:
            xl_op = self._xl_orientation_parameterisations[param_set.xl_ori_param]

        return xl_op

    def _get_xl_unit_cell_parameterisation(self, experiment_id):
        """Return the crystal unit cell parameterisation for the requested
        experiment number (or None if the crystal unit cell in that experiment
        is not parameterised)"""

        param_set = self._exp_to_param[experiment_id]
        xl_ucp = None
        if param_set.xl_uc_param is not None:
            xl_ucp = self._xl_unit_cell_parameterisations[param_set.xl_uc_param]

        return xl_ucp

    def _get_beam_parameterisation(self, experiment_id):
        """Return the beam parameterisation for the requested experiment number
        (or None if the beam in that experiment is not parameterised)"""

        param_set = self._exp_to_param[experiment_id]
        bp = None
        if param_set.beam_param is not None:
            bp = self._beam_parameterisations[param_set.beam_param]

        return bp

    def _get_detector_parameterisation(self, experiment_id):
        """Return the detector parameterisation for the requested experiment number
        (or None if the detector in that experiment is not parameterised)"""

        param_set = self._exp_to_param[experiment_id]
        dp = None
        if param_set.det_param is not None:
            dp = self._detector_parameterisations[param_set.det_param]

        return dp

    def _get_goniometer_parameterisation(self, experiment_id):
        """Return the goniometer parameterisation for the requested experiment number
        (or None if the goniometer in that experiment is not parameterised)"""

        param_set = self._exp_to_param[experiment_id]
        gp = None
        if param_set.gonio_param is not None:
            gp = self._goniometer_parameterisations[param_set.gonio_param]

        return gp

    def _get_state_from_parameterisation(
        self, parameterisation, frame, multi_state_elt=None
    ):
        """Get the model state from the parameterisation at the specified frame,
        taking care of whether it is a scan-varying parameterisation or not"""

        if parameterisation is None:
            return None
        if (
            hasattr(parameterisation, "num_sets")
            and not self._current_frame.get(parameterisation) == frame
        ):
            parameterisation.compose(frame)
            self._current_frame[parameterisation] = frame
        if multi_state_elt is None:
            state = parameterisation.get_state()
        else:
            state = parameterisation.get_state(multi_state_elt=multi_state_elt)
        return state

    def _prepare_for_compose(self, reflections, skip_derivatives=False):
        """Add columns to the reflection table to hold the varying state matrices
        or vectors for the experimental models, if required. Also prepare the cache
        for the derivatives of states that are scan-varying"""

        nref = len(reflections)
        # set columns if needed
        if "u_matrix" not in reflections:
            reflections["u_matrix"] = flex.mat3_double(nref)
        if "b_matrix" not in reflections:
            reflections["b_matrix"] = flex.mat3_double(nref)
        if "s0_vector" not in reflections:
            reflections["s0_vector"] = flex.vec3_double(nref)
        if "d_matrix" not in reflections:
            reflections["d_matrix"] = flex.mat3_double(nref)
        if "D_matrix" not in reflections:
            reflections["D_matrix"] = flex.mat3_double(nref)
        if "S_matrix" not in reflections:
            reflections["S_matrix"] = flex.mat3_double(nref)

        # Clear the state derivative cache and set the number of reflections needed
        # to reconstruct the derivative arrays later
        self._derivative_cache.clear()
        self._derivative_cache.nref = nref

    def compose(self, reflections, skip_derivatives=False):
        """Compose scan-varying crystal parameterisations at the specified image
        number, for the specified experiment, for each image. Put the varying
        matrices in the reflection table, and cache the derivatives."""

        self._prepare_for_compose(reflections, skip_derivatives)

        for iexp, exp in enumerate(self._experiments):

            # select the reflections of interest
            sel = reflections["id"] == iexp
            isel = sel.iselection()

            # skip empty experiments (https://github.com/dials/dials/issues/1417)
            if len(isel) == 0:
                continue

            blocks = reflections["block"].select(isel)

            # identify which parameterisations to use for this experiment
            xl_op = self._get_xl_orientation_parameterisation(iexp)
            xl_ucp = self._get_xl_unit_cell_parameterisation(iexp)
            bp = self._get_beam_parameterisation(iexp)
            dp = self._get_detector_parameterisation(iexp)
            gp = self._get_goniometer_parameterisation(iexp)

            # reset current frame cache for scan-varying parameterisations
            self._current_frame = {}

            # get state and derivatives for each block
            for block in range(flex.min(blocks), flex.max(blocks) + 1):

                # determine the subset of reflections this affects
                subsel = isel.select(blocks == block)
                if len(subsel) == 0:
                    continue

                # get the panels hit by these reflections
                panels = reflections["panel"].select(subsel)

                # get the integer frame number nearest the centre of that block
                frames = reflections["block_centre"].select(subsel)

                # can only be false if original block assignment has gone wrong
                assert frames.all_eq(
                    frames[0]
                ), "Failing: a block contains reflections that shouldn't be there"
                frame = int(math.floor(frames[0]))

                # model states at current frame
                U = self._get_state_from_parameterisation(xl_op, frame)
                if U is None:
                    U = matrix.sqr(exp.crystal.get_U())

                B = self._get_state_from_parameterisation(xl_ucp, frame)
                if B is None:
                    B = matrix.sqr(exp.crystal.get_B())

                s0 = self._get_state_from_parameterisation(bp, frame)
                if s0 is None:
                    s0 = matrix.col(exp.beam.get_s0())

                S = self._get_state_from_parameterisation(gp, frame)
                if S is None:
                    S = matrix.sqr(exp.goniometer.get_setting_rotation())

                # set states for crystal, beam and goniometer
                reflections["u_matrix"].set_selected(subsel, U.elems)
                reflections["b_matrix"].set_selected(subsel, B.elems)
                reflections["s0_vector"].set_selected(subsel, s0.elems)
                reflections["S_matrix"].set_selected(subsel, S.elems)

                # set states and derivatives for this detector
                if dp is not None:  # detector is parameterised
                    if dp.is_multi_state():  # parameterised detector is multi panel

                        # loop through the panels in this detector
                        for panel_id, _ in enumerate(exp.detector):

                            # get the right subset of array indices to set for this panel
                            subsel2 = subsel.select(panels == panel_id)
                            if len(subsel2) == 0:
                                # if no reflections intersect this panel, skip calculation
                                continue

                            dmat = self._get_state_from_parameterisation(
                                dp, frame, multi_state_elt=panel_id
                            )
                            if dmat is None:
                                dmat = exp.detector[panel_id].get_d_matrix()
                            Dmat = exp.detector[panel_id].get_D_matrix()
                            reflections["d_matrix"].set_selected(subsel2, dmat)
                            reflections["D_matrix"].set_selected(subsel2, Dmat)

                            if self._varying_detectors and not skip_derivatives:
                                for j, dd in enumerate(
                                    dp.get_ds_dp(
                                        multi_state_elt=panel_id, use_none_as_null=True
                                    )
                                ):
                                    if dd is None:
                                        continue
                                    self._derivative_cache.append(dp, j, dd, subsel)

                    else:  # parameterised detector is single panel
                        dmat = self._get_state_from_parameterisation(dp, frame)
                        if dmat is None:
                            dmat = exp.detector[0].get_d_matrix()
                        Dmat = exp.detector[0].get_D_matrix()
                        reflections["d_matrix"].set_selected(subsel, dmat)
                        reflections["D_matrix"].set_selected(subsel, Dmat)

                        if self._varying_detectors and not skip_derivatives:
                            for j, dd in enumerate(dp.get_ds_dp(use_none_as_null=True)):
                                if dd is None:
                                    continue
                                self._derivative_cache.append(dp, j, dd, subsel)

                else:  # set states for unparameterised detector (dp is None)
                    # loop through the panels in this detector
                    for panel_id, _ in enumerate(exp.detector):

                        # get the right subset of array indices to set for this panel
                        subsel2 = subsel.select(panels == panel_id)
                        if len(subsel2) == 0:
                            # if no reflections intersect this panel, skip to the next
                            continue

                        dmat = exp.detector[panel_id].get_d_matrix()
                        Dmat = exp.detector[panel_id].get_D_matrix()
                        reflections["d_matrix"].set_selected(subsel2, dmat)
                        reflections["D_matrix"].set_selected(subsel2, Dmat)

                # set derivatives of the states for crystal, beam and goniometer
                if not skip_derivatives:
                    if xl_op is not None and self._varying_xl_orientations:
                        for j, dU in enumerate(xl_op.get_ds_dp(use_none_as_null=True)):
                            if dU is None:
                                continue
                            self._derivative_cache.append(xl_op, j, dU, subsel)
                    if xl_ucp is not None and self._varying_xl_unit_cells:
                        for j, dB in enumerate(xl_ucp.get_ds_dp(use_none_as_null=True)):
                            if dB is None:
                                continue
                            self._derivative_cache.append(xl_ucp, j, dB, subsel)
                    if bp is not None and self._varying_beams:
                        for j, ds0 in enumerate(bp.get_ds_dp(use_none_as_null=True)):
                            if ds0 is None:
                                continue
                            self._derivative_cache.append(bp, j, ds0, subsel)
                    if gp is not None and self._varying_goniometers:
                        for j, dS in enumerate(gp.get_ds_dp(use_none_as_null=True)):
                            if dS is None:
                                continue
                            self._derivative_cache.append(gp, j, dS, subsel)

        # set the UB matrices for prediction
        reflections["ub_matrix"] = reflections["u_matrix"] * reflections["b_matrix"]

    # called by refiner.run for setting the crystal scan points
    def get_varying_UB(self, obs_image_numbers, experiment_id):
        """Extract the setting matrix from the contained scan-dependent crystal
        parameterisations at specified image number."""

        if not (self._varying_xl_unit_cells or self._varying_xl_orientations):
            return None

        # identify which crystal parameterisations to use for this experiment
        xl_op = self._get_xl_orientation_parameterisation(experiment_id)
        xl_ucp = self._get_xl_unit_cell_parameterisation(experiment_id)

        # pre-calculated matrices for fixed parameterisations
        fixed_B = None
        fixed_U = None
        if xl_op is None:
            fixed_U = matrix.sqr(self._experiments[experiment_id].crystal.get_U())
        if xl_ucp is None:
            fixed_B = matrix.sqr(self._experiments[experiment_id].crystal.get_B())

        UB_list = []
        for i in obs_image_numbers:
            if fixed_U:
                U = fixed_U
            else:
                U = self._get_state_from_parameterisation(xl_op, i)
            if fixed_B:
                B = fixed_B
            else:
                B = self._get_state_from_parameterisation(xl_ucp, i)
            UB_list.append(U * B)

        return UB_list

    # called by refiner.run for setting the beam scan points
    def get_varying_s0(self, obs_image_numbers, experiment_id):
        """Extract the s0 vector from the contained scan-dependent beam
        parameterisation at specified image number."""

        if not self._varying_beams:
            return None

        # identify which beam parameterisation to use for this experiment
        bp = self._get_beam_parameterisation(experiment_id)

        s0_list = []
        for i in obs_image_numbers:
            s0 = self._get_state_from_parameterisation(bp, i)
            s0_list.append(s0)

        return s0_list

    # called by refiner.run for setting the goniometer scan points
    def get_varying_setting_rotation(self, obs_image_numbers, experiment_id):
        """Extract the S matrix from the contained scan-dependent goniometer
        parameterisation at specified image number."""

        if not self._varying_goniometers:
            return None

        # identify which goniometer parameterisation to use for this experiment
        gp = self._get_goniometer_parameterisation(experiment_id)

        S_list = []
        for i in obs_image_numbers:
            S = self._get_state_from_parameterisation(gp, i)
            S_list.append(S)

        return S_list

    # overloaded for the scan-varying case
    def _get_model_data_for_experiment(self, experiment, reflections):
        """helper function to return model data s0, U, B, D and S for a particular
        experiment. In this scan-varying overload this is trivial because these
        values are already set as arrays in the reflection table"""

        return {
            "s0": reflections["s0_vector"],
            "U": reflections["u_matrix"],
            "B": reflections["b_matrix"],
            "D": reflections["D_matrix"],
            "S": reflections["S_matrix"],
        }

    def _beam_derivatives(self, isel, parameterisation, reflections):
        """Determine whether ds0_dp was precalculated then call the base class
        method"""

        if self._varying_beams:
            if "imatch" in reflections:
                imatch = reflections["imatch"]
            else:
                imatch = None
            ds0_dxluc_p = self._derivative_cache.build_gradients(
                parameterisation=parameterisation, isel=isel, imatch=imatch
            )
        else:
            ds0_dxluc_p = None

        return super()._beam_derivatives(isel, parameterisation, ds0_dxluc_p)

    def _xl_orientation_derivatives(self, isel, parameterisation, reflections):
        """Determine whether dU_dp was precalculated then call the base class
        method"""

        if self._varying_xl_orientations:
            if "imatch" in reflections:
                imatch = reflections["imatch"]
            else:
                imatch = None
            dU_dxlo_p = self._derivative_cache.build_gradients(
                parameterisation=parameterisation, isel=isel, imatch=imatch
            )
        else:
            dU_dxlo_p = None

        return super()._xl_orientation_derivatives(isel, parameterisation, dU_dxlo_p)

    def _xl_unit_cell_derivatives(self, isel, parameterisation, reflections):
        """Determine whether dB_dp was precalculated then call the base class
        method"""

        if self._varying_xl_unit_cells:
            if "imatch" in reflections:
                imatch = reflections["imatch"]
            else:
                imatch = None
            dB_dxluc_p = self._derivative_cache.build_gradients(
                parameterisation=parameterisation, isel=isel, imatch=imatch
            )
        else:
            dB_dxluc_p = None

        return super()._xl_unit_cell_derivatives(isel, parameterisation, dB_dxluc_p)

    def _detector_derivatives(self, isel, panel_id, parameterisation, reflections):
        """Determine whether dd_dp was precalculated then call the base class
        method"""

        if self._varying_detectors:
            if "imatch" in reflections:
                imatch = reflections["imatch"]
            else:
                imatch = None
            dd_ddet_p = self._derivative_cache.build_gradients(
                parameterisation=parameterisation, isel=isel, imatch=imatch
            )
        else:
            dd_ddet_p = None

        return super()._detector_derivatives(
            isel, panel_id, parameterisation, dd_ddet_p
        )

    def _goniometer_derivatives(self, isel, parameterisation, reflections):
        """Determine whether dS_dp was precalculated then call the base class
        method"""

        if self._varying_goniometers:
            if "imatch" in reflections:
                imatch = reflections["imatch"]
            else:
                imatch = None
            dS_dgon_p = self._derivative_cache.build_gradients(
                parameterisation=parameterisation, isel=isel, imatch=imatch
            )
        else:
            dS_dgon_p = None

        return super()._goniometer_derivatives(isel, parameterisation, dS_dgon_p)

    def calculate_model_state_uncertainties(
        self, var_cov=None, obs_image_number=None, experiment_id=None
    ):
        """Take a variance-covariance matrix of all free parameters (probably
        calculated by a minimisation engine). For each parameterisation in the
        global model, extract the subset of this matrix for the associated block
        of parameters. Pass this on to the relevant model parameterisation to
        calculate its own uncertainty of state.

        This scan-varying version should first be called with var_cov set but
        obs_image_number=None and experiment_id=None. This calls the scan-static
        version to do the calculation for the scan-static parameterisations and
        also caches the subsets of var_cov relevant for the scan-varying
        parameterisations. Subsequent calls should provide obs_image_number and
        experiment_id to calculate for a particular crystal at a particular
        scan-point"""

        # First call, only a variance-covariance matrix is supplied
        if var_cov is not None:
            assert [obs_image_number, experiment_id].count(None) == 2
            super().calculate_model_state_uncertainties(var_cov)
            return

        # Later calls, only an experiment and image number are supplied for
        # identify the crystal parameterisations for this experiment
        xl_op = self._get_xl_orientation_parameterisation(experiment_id)
        xl_ucp = self._get_xl_unit_cell_parameterisation(experiment_id)

        result = {}

        # compose at the requested image number and calculate using the cached
        # varcov matrices. Take the first elt of the list because the crystal
        # parameterisations are not multi-state
        if xl_op is not None:
            try:
                xl_op.compose(obs_image_number)
                result["U_cov"] = xl_op.calculate_state_uncertainties(var_cov=None)[0]
            except TypeError:
                pass

        if xl_ucp is not None:
            try:
                xl_ucp.compose(obs_image_number)
                result["B_cov"] = xl_ucp.calculate_state_uncertainties(var_cov=None)[0]
            except TypeError:
                pass

        return result

    def set_model_state_uncertainties(self, u_cov_list, b_cov_list, experiment_id=None):
        """Identify the crystal parameterisations and set the list of covariance
        matrices, if available. They will only be available if the parameterisation
        is a scan-varying type, otherwise they are None"""

        xl_op = self._get_xl_orientation_parameterisation(experiment_id)
        xl_ucp = self._get_xl_unit_cell_parameterisation(experiment_id)

        if u_cov_list:
            try:
                xl_op.set_state_uncertainties(u_cov_list)
            except AttributeError:
                pass

        if b_cov_list:
            try:
                xl_ucp.set_state_uncertainties(b_cov_list)
            except AttributeError:
                pass


class ScanVaryingPredictionParameterisationSparse(
    SparseGradientVectorMixin, ScanVaryingPredictionParameterisation
):
    pass



"""
Tests for the scaling restraints module.
"""

from __future__ import annotations

from unittest.mock import Mock

import pytest

from scitbx import sparse

from dials.algorithms.scaling.scaling_restraints import (
    ScalingRestraintsCalculator,
    SingleScalingRestraintsCalculator,
)
from dials.array_family import flex


@pytest.fixture
def mock_restrained_component():
    """Mock a component with restraints."""
    component = Mock()
    component.n_params = 3
    component.calculate_restraints.return_value = [
        flex.double([1.0, 2.0, 3.0]),
        flex.double([0.1, 0.2, 0.3]),
    ]
    jacobian_restr = sparse.matrix(component.n_params, component.n_params)
    jacobian_restr[0, 0] = 1.0
    component.calculate_jacobian_restraints.return_value = [
        flex.double([1.0, 1.1, 1.2]),
        jacobian_restr,
        flex.double([1.0, 2.0, 3.0]),
    ]
    return component


@pytest.fixture
def mock_unrestrained_component():
    """Mock a component without restraints."""
    component = Mock()
    component.n_params = 5
    component.calculate_restraints.return_value = None
    component.calculate_jacobian_restraints.return_value = None
    return component


@pytest.fixture
def mock_parameter_manager(mock_restrained_component, mock_unrestrained_component):
    """Mock a parameter manager to handle the components required for the
    ScalingRestraints class."""
    apm = Mock()
    apm.components = {
        "restrained": {
            "object": mock_restrained_component,
            "n_params": mock_restrained_component.n_params,
            "start_idx": 0,
        },
        "unrestrained": {
            "object": mock_unrestrained_component,
            "n_params": mock_unrestrained_component.n_params,
            "start_idx": mock_restrained_component.n_params,
        },
    }
    apm.n_active_params = (
        mock_restrained_component.n_params + mock_unrestrained_component.n_params
    )
    return apm


@pytest.fixture
def mock_unrestrained_apm(mock_unrestrained_component):
    """Mock a parameter manager to handle no restrained components."""
    apm = Mock()
    apm.components = {
        "unrestrained": {
            "object": mock_unrestrained_component,
            "n_params": mock_unrestrained_component.n_params,
            "start_idx": 0,
        }
    }
    apm.n_active_params = mock_unrestrained_component.n_params
    return apm


@pytest.fixture
def mock_multi_apm(mock_parameter_manager):
    """Mock a multi-dataset parameter manager."""
    multi_apm = Mock()
    multi_apm.apm_list = [mock_parameter_manager, mock_parameter_manager]
    n_params = mock_parameter_manager.n_active_params
    multi_apm.n_active_params = n_params * 2
    multi_apm.apm_data = {
        0: {"start_idx": 0, "end_idx": n_params},
        1: {"start_idx": n_params, "end_idx": n_params * 2},
    }
    return multi_apm


@pytest.fixture
def mock_multi_unrestrained_apm(mock_unrestrained_apm):
    """Mock a parameter manager to handle no restrained components."""
    multi_apm = Mock()
    multi_apm.apm_list = [mock_unrestrained_apm, mock_unrestrained_apm]
    return multi_apm


def test_unrestrained_ScalingRestraints(
    mock_unrestrained_apm, mock_multi_unrestrained_apm
):
    """Test the case of unrestrained components. None should be returned in each
    case."""

    assert (
        SingleScalingRestraintsCalculator.calculate_restraints(mock_unrestrained_apm)
        is None
    )
    assert (
        SingleScalingRestraintsCalculator.calculate_jacobian_restraints(
            mock_unrestrained_apm
        )
        is None
    )

    assert (
        ScalingRestraintsCalculator.calculate_restraints(mock_multi_unrestrained_apm)
        is None
    )
    assert (
        ScalingRestraintsCalculator.calculate_jacobian_restraints(
            mock_multi_unrestrained_apm
        )
        is None
    )


def test_ScalingRestraints(
    mock_parameter_manager, mock_restrained_component, mock_unrestrained_component
):
    """Test for the single scaling restraints manager."""

    # Test the call to calculate restraints. This should return a residual
    # vector of the same length as the restraints of the restrained component,
    # and a gradient vector of the total length of all parameters.
    restraints = SingleScalingRestraintsCalculator.calculate_restraints(
        mock_parameter_manager
    )
    abs_restraints = mock_restrained_component.calculate_restraints()
    assert list(restraints[0]) == list(abs_restraints[0])
    assert list(restraints[1]) == (
        list(abs_restraints[1]) + [0.0] * mock_unrestrained_component.n_params
    )

    # Test the call to calculate jacobian restraints. This should return a
    # restraints vector, jacobian and weights vector from the components. The
    # restraints and weights should be the length of the restrained component.
    # The jacobian has n_rows equal to the number of restrainted parameters,
    # n_cols equal to the total number of parameters. Check that these are
    # correctly composed.
    jacobian_restraints = (
        SingleScalingRestraintsCalculator.calculate_jacobian_restraints(
            mock_parameter_manager
        )
    )
    abs_restraints = mock_restrained_component.calculate_jacobian_restraints()
    assert list(jacobian_restraints[0]) == list(abs_restraints[0])
    assert jacobian_restraints[1].n_cols == mock_parameter_manager.n_active_params
    assert jacobian_restraints[1].n_rows == mock_restrained_component.n_params
    for i in range(mock_restrained_component.n_params):
        for j in range(mock_restrained_component.n_params):
            assert jacobian_restraints[1][i, j] == abs_restraints[1][i, j]
    # All other elements should be zero
    assert abs_restraints[1].non_zeroes == jacobian_restraints[1].non_zeroes


def test_MultiScalingRestraints(
    mock_multi_apm, mock_restrained_component, mock_unrestrained_component
):
    """Test for the multi-dataset scaling restraints manager."""

    # Test the call to calculate restraints. Expected return is the individual
    # dataset vectors joined together.
    restraints = ScalingRestraintsCalculator.calculate_restraints(mock_multi_apm)
    abs_restraints = mock_restrained_component.calculate_restraints()
    assert list(restraints[0]) == (list(abs_restraints[0]) + list(abs_restraints[0]))

    assert list(restraints[1]) == (
        list(abs_restraints[1])
        + [0.0] * mock_unrestrained_component.n_params
        + list(abs_restraints[1])
        + [0.0] * mock_unrestrained_component.n_params
    )

    # Test the call to calculate jacobian restraints. Again, the expected return
    # is the individual dataset vectors joined together.
    jacobian_restraints = ScalingRestraintsCalculator.calculate_jacobian_restraints(
        mock_multi_apm
    )
    abs_restraints = mock_restrained_component.calculate_jacobian_restraints()

    assert list(jacobian_restraints[0]) == (
        list(abs_restraints[0]) + list(abs_restraints[0])
    )
    n_abs_params = mock_restrained_component.n_params
    n_total_params = mock_multi_apm.apm_data[0]["end_idx"]
    assert jacobian_restraints[1].n_cols == mock_multi_apm.n_active_params
    assert jacobian_restraints[1].n_rows == n_abs_params * 2
    # Check that both restraints jacobians were set in correct location.
    for i in range(mock_restrained_component.n_params):
        for j in range(mock_restrained_component.n_params):
            assert jacobian_restraints[1][i, j] == abs_restraints[1][i, j]
            assert jacobian_restraints[1][i + n_abs_params, j + n_total_params] == (
                abs_restraints[1][i, j]
            )
    assert abs_restraints[1].non_zeroes * 2 == jacobian_restraints[1].non_zeroes



from __future__ import annotations

import copy
import os

import pytest

from .test_scan_static_reflection_predictor import (  # noqa: F401, used as test fixture; isort:skip
    data as static_test,
)


class Data:
    def __init__(self, dials_regression):
        path = os.path.join(
            dials_regression,
            "prediction_test_data",
            "experiments_scan_varying_crystal.json",
        )

        from dxtbx.model.experiment_list import ExperimentListFactory

        self.experiments = ExperimentListFactory.from_json_file(path)
        assert len(self.experiments) == 1
        assert (
            self.experiments[0].crystal.num_scan_points
            == self.experiments[0].scan.get_num_images() + 1
        )

    def run(self):
        self.tst_regression()
        # self.tst_with_hkl()
        # self.tst_with_hkl_and_panel()
        # self.tst_with_hkl_and_panel_list()
        self.tst_for_reflection_table()
        self.tst_vs_static()

    def _predict_new(self, hkl=None, frame=None, panel=None):
        from dials.algorithms.spot_prediction import ScanVaryingReflectionPredictor
        from dials.array_family import flex

        predict = ScanVaryingReflectionPredictor(self.experiments[0])
        # if hkl is None:
        A = [
            self.experiments[0].crystal.get_A_at_scan_point(i)
            for i in range(self.experiments[0].crystal.num_scan_points)
        ]
        result = predict.for_ub(flex.mat3_double(A))
        # else:
        # if panel is None:
        # result = predict(hkl, frame)
        # else:
        # result = predict(hkl, frame, panel)
        return result


@pytest.fixture(scope="session")
def data(dials_regression):
    return Data(dials_regression)


def test_regression(data):
    r_new = data._predict_new()
    assert len(r_new) == 1934
    assert r_new[0]["miller_index"] == (-8, -30, -23)
    assert r_new[0]["xyzcal.px"] == pytest.approx(
        (75.33831543451907, 2327.55737978813, 0.2548567552525226), abs=1e-6
    )

    # def tst_with_hkl(self):
    # from dials.algorithms.spot_prediction import ReekeIndexGenerator
    # from dials.array_family import flex
    # from scitbx import matrix
    # import scitbx

    # m2 = self.experiments[0].goniometer.get_rotation_axis()
    # s0 = self.experiments[0].beam.get_s0()
    # dmin = self.experiments[0].detector.get_max_resolution(s0)
    # margin = 1
    # scan = self.experiments[0].scan
    # crystal = self.experiments[0].crystal
    # frame_0 = scan.get_array_range()[0]
    # step = 1

    # all_indices = flex.miller_index()
    # all_frames = flex.int()
    # for frame in range(*scan.get_array_range()):

    # phi_beg = scan.get_angle_from_array_index(frame, deg = False)
    # phi_end = scan.get_angle_from_array_index(frame + step, deg = False)
    # r_beg = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_beg, deg = False))
    # r_end = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_end, deg = False))

    # A1 = r_beg * crystal.get_A_at_scan_point(frame - frame_0)

    # A2 = r_end * crystal.get_A_at_scan_point(frame - frame_0 + step)

    # indices = ReekeIndexGenerator(A1, A2, m2, s0, dmin, margin)
    # indices = indices.to_array()
    # all_indices.extend(indices)
    # all_frames.extend(flex.int(len(indices), frame))

    # r_old = self._predict_new()
    # r_new = self._predict_new(all_indices, all_frames)
    # assert(len(r_old) == len(r_new))
    # print 'OK'

    # def tst_with_hkl_and_panel(self):
    # from dials.algorithms.spot_prediction import ReekeIndexGenerator
    # from dials.array_family import flex
    # from scitbx import matrix
    # import scitbx

    # m2 = self.experiments[0].goniometer.get_rotation_axis()
    # s0 = self.experiments[0].beam.get_s0()
    # dmin = self.experiments[0].detector.get_max_resolution(s0)
    # margin = 1
    # scan = self.experiments[0].scan
    # crystal = self.experiments[0].crystal
    # frame_0 = scan.get_array_range()[0]
    # step = 1

    # all_indices = flex.miller_index()
    # all_frames = flex.int()
    # for frame in range(*scan.get_array_range()):

    # phi_beg = scan.get_angle_from_array_index(frame, deg = False)
    # phi_end = scan.get_angle_from_array_index(frame + step, deg = False)
    # r_beg = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_beg, deg = False))
    # r_end = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_end, deg = False))

    # A1 = r_beg * crystal.get_A_at_scan_point(frame - frame_0)

    # A2 = r_end * crystal.get_A_at_scan_point(frame - frame_0 + step)

    # indices = ReekeIndexGenerator(A1, A2, m2, s0, dmin, margin)
    # indices = indices.to_array()
    # all_indices.extend(indices)
    # all_frames.extend(flex.int(len(indices), frame))

    # r_old = self._predict_new()
    # try:
    # r_new = self._predict_new(all_indices, all_frames, 1)
    # assert(False)
    # except Exception:
    # pass

    # r_new = self._predict_new(all_indices, all_frames, 0)
    # assert(len(r_old) < len(r_new))
    # print 'OK'

    # def tst_with_hkl_and_panel_list(self):

    # from dials.algorithms.spot_prediction import ReekeIndexGenerator
    # from dials.array_family import flex
    # from scitbx import matrix
    # import scitbx

    # m2 = self.experiments[0].goniometer.get_rotation_axis()
    # s0 = self.experiments[0].beam.get_s0()
    # dmin = self.experiments[0].detector.get_max_resolution(s0)
    # margin = 1
    # scan = self.experiments[0].scan
    # crystal = self.experiments[0].crystal
    # frame_0 = scan.get_array_range()[0]
    # step = 1

    # all_indices = flex.miller_index()
    # all_frames = flex.int()
    # for frame in range(*scan.get_array_range()):

    # phi_beg = scan.get_angle_from_array_index(frame, deg = False)
    # phi_end = scan.get_angle_from_array_index(frame + step, deg = False)
    # r_beg = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_beg, deg = False))
    # r_end = matrix.sqr(scitbx.math.r3_rotation_axis_and_angle_as_matrix(
    # axis = m2, angle = phi_end, deg = False))

    # A1 = r_beg * crystal.get_A_at_scan_point(frame - frame_0)

    # A2 = r_end * crystal.get_A_at_scan_point(frame - frame_0 + step)

    # indices = ReekeIndexGenerator(A1, A2, m2, s0, dmin, margin)
    # indices = indices.to_array()
    # all_indices.extend(indices)
    # all_frames.extend(flex.int(len(indices), frame))

    # r_old = self._predict_new()
    # try:
    # r_new = self._predict_new(all_indices, all_frames,
    # flex.size_t(len(all_indices), 1))
    # assert(False)
    # except Exception:
    # pass

    # r_new = self._predict_new(all_indices, all_frames,
    # flex.size_t(len(all_indices), 0))
    # assert(len(r_old) < len(r_new))
    # print 'OK'


def test_scan_varying_results_are_close_to_static_prediction_when_model_is_static(
    static_test,  # noqa: F811, not a redefinition
):
    """Test that various modes of scan-varying prediction produce results
    close to static prediction when the supplied models are indeed static"""

    # Get static predictor results

    scan = static_test.experiments[0].scan
    crystal = static_test.experiments[0].crystal
    beam = static_test.experiments[0].beam
    goniometer = static_test.experiments[0].goniometer
    n_scan_points = scan.get_num_images() + 1
    static_preds = static_test.predict_new()
    static_preds.sort("miller_index")

    # Set up scan-varying predictor
    from dials.algorithms.spot_prediction import ScanVaryingReflectionPredictor
    from dials.array_family import flex

    predict = ScanVaryingReflectionPredictor(static_test.experiments[0])

    def compare(refs1, refs2):
        assert len(refs1) == len(refs2)
        for r1, r2 in zip(refs1.rows(), refs2.rows()):
            assert r1["miller_index"] == r2["miller_index"]
            # differences less than one hundredth of a pixel/image
            for e1, e2 in zip(r1["xyzcal.px"], r2["xyzcal.px"]):
                assert e1 == pytest.approx(e2, abs=0.01)

    # Prediction for UB matrix expressed as array of static UB
    A = [crystal.get_A() for i in range(n_scan_points)]
    result1 = predict.for_ub(flex.mat3_double(A))
    result1.sort("miller_index")
    compare(static_preds, result1)

    # Prediction for UB matrix, s0 vectors and goniometer setting rotation
    # matrices expressed as arrays of static model states
    s0 = [beam.get_s0() for i in range(n_scan_points)]
    S = [goniometer.get_setting_rotation() for i in range(n_scan_points)]
    result2 = predict.for_varying_models(
        flex.mat3_double(A), flex.vec3_double(s0), flex.mat3_double(S)
    )
    result2.sort("miller_index")
    compare(static_preds, result2)

    # First frame only, start and end UB
    _, _, z = static_preds["xyzcal.px"].parts()
    static_preds_frame0 = static_preds.select((z >= 0) & (z < 1))
    A = crystal.get_A()
    result3 = predict.for_ub_on_single_image(0, A, A)
    result3.sort("miller_index")
    compare(static_preds_frame0, result3)

    # First frame only, start and end UB, s0 and S
    s0 = beam.get_s0()
    S = goniometer.get_setting_rotation()
    result4 = predict.for_varying_models_on_single_image(0, A, A, s0, s0, S, S)
    result4.sort("miller_index")
    compare(static_preds_frame0, result4)


def test_for_reflection_table(data):
    from dials.algorithms.spot_prediction import (
        ScanStaticReflectionPredictor,
        ScanVaryingReflectionPredictor,
    )
    from dials.array_family import flex

    predict = ScanStaticReflectionPredictor(data.experiments[0])
    preds = predict.for_ub(data.experiments[0].crystal.get_A())

    preds["ub_matrix"] = flex.mat3_double(
        len(preds), data.experiments[0].crystal.get_A()
    )
    preds["s0"] = flex.vec3_double(len(preds), data.experiments[0].beam.get_s0())
    preds["d_matrix"] = flex.mat3_double(len(preds))
    preds["S_matrix"] = flex.mat3_double(
        len(preds), data.experiments[0].goniometer.get_setting_rotation()
    )
    for ipanel, panel in enumerate(data.experiments[0].detector):
        sel = preds["panel"] == ipanel
        D = panel.get_d_matrix()
        preds["d_matrix"].set_selected(sel, D)
    predict = ScanVaryingReflectionPredictor(data.experiments[0])
    old_preds = copy.deepcopy(preds)
    predict.for_reflection_table(
        preds, preds["ub_matrix"], preds["s0"], preds["d_matrix"], preds["S_matrix"]
    )

    # Because UB, s0, d and S values are the same for all reflections, the new
    # reflections should be approx equal to those produced by the scan static
    # predictor
    old_x, old_y, old_z = old_preds["xyzcal.px"].parts()
    new_x, new_y, new_z = preds["xyzcal.px"].parts()
    assert old_x.all_approx_equal(new_x)
    assert old_y.all_approx_equal(new_y)
    assert old_z.all_approx_equal(new_z)



from __future__ import annotations

import random

import pytest


def test_mean_filter():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import mean_filter

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    mean = mean_filter(image, (3, 3))

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    eps = 1e-7
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        m1 = mean[j, i]
        p = image[j - 3 : j + 4, i - 3 : i + 4]
        mv = flex.mean_and_variance(p.as_1d())
        m2 = mv.mean()
        assert m1 == pytest.approx(m2, abs=eps)


def test_masked_mean_filter():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import mean_filter

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))
    mask = flex.random_bool(2000 * 2000, 0.99).as_int()
    mask.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    mask2 = mask.deep_copy()
    mean = mean_filter(image, mask2, (3, 3), 1)

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    eps = 1e-7
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        m1 = mean[j, i]
        p = image[j - 3 : j + 4, i - 3 : i + 4]
        m = mask[j - 3 : j + 4, i - 3 : i + 4]
        if mask[j, i] == 0:
            m2 = 0.0
        else:
            p = flex.select(p, flags=m)
            mv = flex.mean_and_variance(flex.double(p))
            m2 = mv.mean()
        assert m1 == pytest.approx(m2, abs=eps)


def test_mean_and_variance_filter():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import mean_and_variance_filter

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    mean_and_variance = mean_and_variance_filter(image, (3, 3))
    mean = mean_and_variance.mean()
    sample_variance = mean_and_variance.sample_variance()

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    eps = 1e-7
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        m1 = mean[j, i]
        sv1 = sample_variance[j, i]
        p = image[j - 3 : j + 4, i - 3 : i + 4]
        mv = flex.mean_and_variance(p.as_1d())
        m2 = mv.mean()
        sv2 = mv.unweighted_sample_variance()
        assert m1 == pytest.approx(m2, abs=eps)
        assert sv1 == pytest.approx(sv2, abs=eps)


def test_masked_mean_and_variance_filter():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import mean_and_variance_filter

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))
    mask = flex.random_bool(2000 * 2000, 0.99).as_int()
    mask.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    mask2 = mask.deep_copy()
    mv = mean_and_variance_filter(image, mask2, (3, 3), 2)
    mean = mv.mean()
    var = mv.sample_variance()

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    eps = 1e-7
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        m1 = mean[j, i]
        v1 = var[j, i]
        p = image[j - 3 : j + 4, i - 3 : i + 4]
        m = mask[j - 3 : j + 4, i - 3 : i + 4]
        if mask[j, i] == 0:
            m2 = 0.0
            v2 = 0.0
        else:
            p = flex.select(p, flags=m)
            mv = flex.mean_and_variance(flex.double(p))
            m2 = mv.mean()
            v2 = mv.unweighted_sample_variance()
        assert m1 == pytest.approx(m2, abs=eps)
        assert v1 == pytest.approx(v2, abs=eps)



from __future__ import annotations

import pytest

from cctbx.sgtbx import space_group, space_group_symbols
from dxtbx.model import ScanFactory
from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse
from scitbx.math import five_number_summary

from dials.algorithms.refinement.parameterisation.beam_parameters import (
    BeamParameterisation,
)
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.parameterisation.prediction_parameters_stills import (
    SphericalRelpStillsPredictionParameterisation,
    StillsPredictionParameterisation,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    ExperimentsPredictorFactory,
    ScansRayPredictor,
    StillsExperimentsPredictor,
)
from dials.algorithms.spot_prediction import IndexGenerator
from dials.array_family import flex

from . import geometry_phil
from .setup_geometry import Extract


class _Test:
    def __init__(self):
        self.create_models()
        self.generate_reflections()

    def create_models(self):
        # build models, with a larger crystal than default in order to get plenty of
        # reflections on the 'still' image
        overrides = """
    geometry.parameters.crystal.a.length.range=40 50;
    geometry.parameters.crystal.b.length.range=40 50;
    geometry.parameters.crystal.c.length.range=40 50;
    geometry.parameters.random_seed = 42"""

        master_phil = parse(geometry_phil)

        models = Extract(master_phil, overrides)

        # keep track of the models
        self.detector = models.detector
        self.gonio = models.goniometer
        self.crystal = models.crystal
        self.beam = models.beam

        # Create a stills ExperimentList
        self.stills_experiments = ExperimentList()
        self.stills_experiments.append(
            Experiment(
                beam=self.beam,
                detector=self.detector,
                crystal=self.crystal,
                imageset=None,
            )
        )

        # keep track of the parameterisation of the models
        self.det_param = DetectorParameterisationSinglePanel(self.detector)
        self.s0_param = BeamParameterisation(self.beam, self.gonio)
        self.xlo_param = CrystalOrientationParameterisation(self.crystal)
        self.xluc_param = CrystalUnitCellParameterisation(self.crystal)

    def generate_reflections(self):
        # Build a mock scan for a 3 degree sequence
        sf = ScanFactory()
        self.scan = sf.make_scan(
            image_range=(1, 1),
            exposure_times=0.1,
            oscillation=(0, 3.0),
            epochs=list(range(1)),
            deg=True,
        )
        sequence_range = self.scan.get_oscillation_range(deg=False)

        # Create a scans ExperimentList, only for generating reflections
        experiments = ExperimentList()
        experiments.append(
            Experiment(
                beam=self.beam,
                detector=self.detector,
                goniometer=self.gonio,
                scan=self.scan,
                crystal=self.crystal,
                imageset=None,
            )
        )

        # Create a ScansRayPredictor
        ray_predictor = ScansRayPredictor(experiments, sequence_range)

        # Generate rays - only to work out which hkls are predicted
        resolution = 2.0
        index_generator = IndexGenerator(
            self.crystal.get_unit_cell(),
            space_group(space_group_symbols(1).hall()).type(),
            resolution,
        )
        indices = index_generator.to_array()
        rays = ray_predictor(indices)

        # Make a standard reflection_table and copy in the ray data
        self.reflections = flex.reflection_table.empty_standard(len(rays))
        self.reflections.update(rays)

        # Set dummy observed variances to allow statistical weights to be set
        self.reflections["xyzobs.mm.variance"] += (1e-3, 1e-3, 1e-6)

    def get_fd_gradients(self, pred_param, ref_predictor):

        # get finite difference gradients
        p_vals = pred_param.get_param_vals()
        deltas = [1.0e-7] * len(p_vals)

        fd_grads = []
        p_names = pred_param.get_param_names()
        for i in range(len(deltas)):

            # save parameter value
            val = p_vals[i]

            # calc reverse state
            p_vals[i] -= deltas[i] / 2.0
            pred_param.set_param_vals(p_vals)

            ref_predictor(self.reflections)

            x, y, _ = self.reflections["xyzcal.mm"].deep_copy().parts()
            delpsi = self.reflections["delpsical.rad"].deep_copy()
            rev_state = flex.vec3_double(x, y, delpsi)

            # calc forward state
            p_vals[i] += deltas[i]
            pred_param.set_param_vals(p_vals)

            ref_predictor(self.reflections)

            x, y, _ = self.reflections["xyzcal.mm"].deep_copy().parts()
            delpsi = self.reflections["delpsical.rad"].deep_copy()
            fwd_state = flex.vec3_double(x, y, delpsi)

            # reset parameter to saved value
            p_vals[i] = val

            # finite difference
            fd = fwd_state - rev_state
            x_grads, y_grads, delpsi_grads = fd.parts()
            x_grads /= deltas[i]
            y_grads /= deltas[i]
            delpsi_grads /= deltas[i]

            fd_grads.append(
                {
                    "name": p_names[i],
                    "dX_dp": x_grads,
                    "dY_dp": y_grads,
                    "dDeltaPsi_dp": delpsi_grads,
                }
            )

        # return to the initial state
        pred_param.set_param_vals(p_vals)

        return fd_grads


@pytest.fixture(scope="session")
def tc():
    return _Test()


def test_stills_pred_param(tc):
    print("Testing derivatives for StillsPredictionParameterisation")
    print("========================================================")

    # Build a prediction parameterisation for the stills experiment
    pred_param = StillsPredictionParameterisation(
        tc.stills_experiments,
        detector_parameterisations=[tc.det_param],
        beam_parameterisations=[tc.s0_param],
        xl_orientation_parameterisations=[tc.xlo_param],
        xl_unit_cell_parameterisations=[tc.xluc_param],
    )

    # Predict the reflections in place. Must do this ahead of calculating
    # the analytical gradients so quantities like s1 are correct

    ref_predictor = StillsExperimentsPredictor(tc.stills_experiments)
    ref_predictor(tc.reflections)

    # get analytical gradients
    an_grads = pred_param.get_gradients(tc.reflections)

    fd_grads = tc.get_fd_gradients(pred_param, ref_predictor)

    for i, (an_grad, fd_grad) in enumerate(zip(an_grads, fd_grads)):

        # compare FD with analytical calculations
        print(f"\nParameter {i}: {fd_grad['name']}")

        for name in ["dX_dp", "dY_dp", "dDeltaPsi_dp"]:
            print(name)
            a = fd_grad[name]
            b = an_grad[name]

            abs_error = a - b

            fns = five_number_summary(abs_error)
            print(
                ("  summary of absolute errors: %9.6f %9.6f %9.6f " + "%9.6f %9.6f")
                % fns
            )
            assert flex.max(flex.abs(abs_error)) < 0.0003
            # largest absolute error found to be about 0.00025 for dY/dp of
            # Crystal0g_param_3. Reject outlying absolute errors and test again.
            iqr = fns[3] - fns[1]

            # skip further stats on errors with an iqr of near zero, e.g. dDeltaPsi_dp
            # for detector parameters, which are all equal to zero
            if iqr < 1.0e-10:
                continue

            sel1 = abs_error < fns[3] + 1.5 * iqr
            sel2 = abs_error > fns[1] - 1.5 * iqr
            sel = sel1 & sel2
            tst = flex.max_index(flex.abs(abs_error.select(sel)))
            tst_val = abs_error.select(sel)[tst]
            n_outliers = sel.count(False)
            print(
                (
                    "  {0} outliers rejected, leaving greatest "
                    + "absolute error: {1:9.6f}"
                ).format(n_outliers, tst_val)
            )
            # largest absolute error now 0.000086 for dX/dp of Beam0Mu2
            assert abs(tst_val) < 0.00009

            # Completely skip parameters with FD gradients all zero (e.g. gradients of
            # DeltaPsi for detector parameters)
            sel1 = flex.abs(a) < 1.0e-10
            if sel1.all_eq(True):
                continue

            # otherwise calculate normalised errors, by dividing absolute errors by
            # the IQR (more stable than relative error calculation)
            norm_error = abs_error / iqr
            fns = five_number_summary(norm_error)
            print(
                ("  summary of normalised errors: %9.6f %9.6f %9.6f " + "%9.6f %9.6f")
                % fns
            )
            # largest normalised error found to be about 25.7 for dY/dp of
            # Crystal0g_param_3.
            try:
                assert flex.max(flex.abs(norm_error)) < 30
            except AssertionError as e:
                e.args += (
                    f"extreme normalised error value: {flex.max(flex.abs(norm_error))}",
                )
                raise e

            # Reject outlying normalised errors and test again
            iqr = fns[3] - fns[1]
            if iqr > 0.0:
                sel1 = norm_error < fns[3] + 1.5 * iqr
                sel2 = norm_error > fns[1] - 1.5 * iqr
                sel = sel1 & sel2
                tst = flex.max_index(flex.abs(norm_error.select(sel)))
                tst_val = norm_error.select(sel)[tst]
                n_outliers = sel.count(False)

                # most outliers found for for dY/dp of Crystal0g_param_3 (which had
                # largest errors, so no surprise there).
                try:
                    assert n_outliers < 250
                except AssertionError as e:
                    e.args += (f"too many outliers rejected: {n_outliers}",)
                    raise e

                print(
                    (
                        "  {0} outliers rejected, leaving greatest "
                        + "normalised error: {1:9.6f}"
                    ).format(n_outliers, tst_val)
                )
                # largest normalied error now about -4. for dX/dp of Detector0Tau1
                assert abs(tst_val) < 6, f"should be < 6, not {tst_val}"


# In comparison with FD approximations, the worst gradients by far are dX/dp
# and dY/dp for parameter Crystal0g_param_3. Is this to do with the geometry
# of the test case?


def test_spherical_relp_stills_pred_param(tc):
    print("Testing derivatives for SphericalRelpStillsPredictionParameterisation")
    print("=====================================================================")

    # Build a prediction parameterisation for the stills experiment
    pred_param = SphericalRelpStillsPredictionParameterisation(
        tc.stills_experiments,
        detector_parameterisations=[tc.det_param],
        beam_parameterisations=[tc.s0_param],
        xl_orientation_parameterisations=[tc.xlo_param],
        xl_unit_cell_parameterisations=[tc.xluc_param],
    )

    # Predict the reflections in place. Must do this ahead of calculating
    # the analytical gradients so quantities like s1 are correct
    ref_predictor = ExperimentsPredictorFactory.from_experiments(
        tc.stills_experiments, force_stills=True, spherical_relp=True
    )
    ref_predictor(tc.reflections)

    # get analytical gradients
    an_grads = pred_param.get_gradients(tc.reflections)

    fd_grads = tc.get_fd_gradients(pred_param, ref_predictor)

    # compare FD with analytical calculations
    for i, (an_grad, fd_grad) in enumerate(zip(an_grads, fd_grads)):
        print(f"\nParameter {i}: {fd_grad['name']}")
        for name in ["dX_dp", "dY_dp", "dDeltaPsi_dp"]:
            print(name)
            for a, b in zip(an_grad[name], fd_grad[name]):
                if name == "dDeltaPsi_dp":
                    # DeltaPsi errors are much worse than X, Y errors!
                    # FIXME, look into this further
                    assert a == pytest.approx(b, abs=5e-3)
                else:
                    assert a == pytest.approx(b, abs=5e-6)
            print("OK")



from __future__ import annotations

import logging
import math
import pickle

from iotbx import phil
from scitbx import matrix

from dials.array_family import flex
from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.options import ArgumentParser, flatten_experiments
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.generate_distortion_maps")

help_message = """

Generate dx.pickle, dy.pickle distortion maps for a detector model picked up
from an image file or experiment.expt. These maps can be used to
represent distortion within the millimetre to pixel mapping

Examples::

  dials.generate_distortion_maps image_001.cbf dx=0.5 dy=1.5
  dials.generate_distortion_maps models.expt mode=ellipse phi=0 l2=0.95
"""

scope = phil.parse(
    """
  mode = *translate ellipse
    .type = choice

  translate
    .help = "Options for applying separate translation offsets to each panel"
  {
    dx = 0.0
      .type = floats
      .help = "Translation in pixels to be applied along the fast direction."
              "The number of values supplied should equal the number of panels."
    dy = 0.0
      .type = floats
      .help = "Translation in pixels to be applied along the slow direction."
              "The number of values supplied should equal the number of panels."
  }

  ellipse
    .help = "Options for correcting for elliptical distortion of images."
            "Defaults set for correction of datasets published in"
            "https://doi.org/10.1107/S2059798317010348"
  {
    phi = -21.0
      .type = float
      .help = "Acute angle of one principal axis of the ellipse from the fast"
              "axis of the first panel of the detector"
    l1 = 1.0
      .type = float
      .help = "Scale factor for first axis of the ellipse"

    l2 = 0.956
      .type = float
      .help = "Scale factor for second axis of the ellipse"

    centre_xy = 33.2475,33.2475
      .type = floats(size=2)
      .help = "Centre of the ellipse in millimetres along fast, slow of the"
              "first panel"
  }

  output {
    x_map = dx.pickle
      .type = str
    y_map = dy.pickle
      .type = str
    log = dials.generate_distortion_maps.log
      .type = str
  }
"""
)


def make_dx_dy_translate(imageset, dx, dy):
    images = imageset.indices()
    image = imageset[images[0]]

    if (len(dx) != len(image)) or (len(dx) != len(image)):
        raise Sorry(
            "Please provide separate translations for each panel of the detector"
        )

    distortion_map_x = []
    distortion_map_y = []

    for block, shift_x, shift_y in zip(image, dx, dy):
        distortion_map_x.append(flex.double(flex.grid(block.focus()), shift_x))
        distortion_map_y.append(flex.double(flex.grid(block.focus()), shift_y))

    return tuple(distortion_map_x), tuple(distortion_map_y)


def ellipse_matrix_form(phi, l1, l2):
    """Return the matrix for the quadratic form describing the oblique ellipse
    where the first axis makes an angle phi with the X axis and the scale factors
    for the axes are l1 and l2.
    See https://www.le.ac.uk/users/dsgp1/COURSES/TOPICS/quadrat.pdf"""
    deg2rad = math.pi / 180.0
    phi *= deg2rad
    cphi = math.cos(phi)
    sphi = math.sin(phi)

    a11 = l1 * cphi**2 + l2 * sphi**2
    a12 = (l2 - l1) * sphi * cphi
    a21 = a12
    a22 = l1 * sphi**2 + l2 * cphi**2

    assert a11 * a22 - 2 * a12 > 0.0

    return matrix.sqr((a11, a12, a21, a22))


def make_dx_dy_ellipse(imageset, phi, l1, l2, centre_xy):
    detector = imageset.get_detector()

    # Get fast and slow axes from the first panel. These will form the X and Y
    # directions for the Cartesian coordinates of the correction map
    p0 = detector[0]
    f0, s0 = matrix.col(p0.get_fast_axis()), matrix.col(p0.get_slow_axis())

    # Get the lab coordinate of the centre of the ellipse
    topleft = matrix.col(p0.get_pixel_lab_coord((0, 0)))
    mid = topleft + centre_xy[0] * f0 + centre_xy[1] * s0

    # Get matrix describing the elliptical distortion
    M = ellipse_matrix_form(phi, l1, l2)

    distortion_map_x = []
    distortion_map_y = []

    for panel in detector:
        size_x, size_y = panel.get_pixel_size()
        nx, ny = panel.get_image_size()
        dx = flex.double(flex.grid(nx, ny), 0.0)
        dy = flex.double(flex.grid(nx, ny), 0.0)
        elt = 0
        for j in range(ny):
            for i in range(nx):
                # Get the X,Y coordinate of this pixel in the frame of the correction
                # map, which has its origin at the centre of the ellipse and is aligned
                # along fast, slow of the first panel.
                lab = matrix.col(panel.get_pixel_lab_coord((i + 0.5, j + 0.5)))
                offset = lab - mid
                x = offset.dot(f0)  # undistorted X coordinate (mm)
                y = offset.dot(s0)  # undistorted Y coordinate (mm)
                distort = M * matrix.col((x, y))  # distorted by ellipse matrix

                # store correction in units of the pixel size
                dx[elt] = (x - distort[0]) / size_x
                dy[elt] = (y - distort[1]) / size_y
                elt += 1
        # Add results for this panel
        distortion_map_x.append(dx)
        distortion_map_y.append(dy)

    distortion_map_x = tuple(distortion_map_x)
    distortion_map_y = tuple(distortion_map_y)

    return distortion_map_x, distortion_map_y


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.generate_distortion_maps [options] image_*.cbf"

    parser = ArgumentParser(
        usage=usage,
        phil=scope,
        read_experiments=True,
        read_experiments_from_images=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args)

    # Configure the logging
    log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        exit()

    assert len(experiments) == 1

    imagesets = experiments.imagesets()

    assert len(imagesets) == 1

    imageset = imagesets[0]

    if params.mode == "translate":
        op = params.translate
        logger.info(f"Generating translation map with dx={op.dx}, dy={op.dy}")
        dx, dy = make_dx_dy_translate(imageset, op.dx, op.dy)
    elif params.mode == "ellipse":
        op = params.ellipse
        logger.info(
            "Generating elliptical map with phi={}, l1={}, "
            "l2={}, centre_xy={},{}".format(op.phi, op.l1, op.l2, *op.centre_xy)
        )
        dx, dy = make_dx_dy_ellipse(imageset, op.phi, op.l1, op.l2, op.centre_xy)
    else:
        raise Sorry("Unrecognised mode")

    logger.info(f"Saving X distortion map to {params.output.x_map}")
    with open(params.output.x_map, "wb") as f:
        pickle.dump(dx, f, pickle.HIGHEST_PROTOCOL)

    logger.info(f"Saving Y distortion map to {params.output.y_map}")
    with open(params.output.y_map, "wb") as f:
        pickle.dump(dy, f, pickle.HIGHEST_PROTOCOL)


if __name__ == "__main__":
    run()



"""Functions to help with reindexing against a reference dataset."""

from __future__ import annotations

from cctbx import sgtbx
from mmtbx.scaling.twin_analyses import twin_laws

import dials.util
from dials.util import Sorry


def determine_reindex_operator_against_reference(test_miller_set, reference_miller_set):
    """Reindex a miller set to match a reference miller set.

    This function takes two miller arrays, a reference and a test array. The
    space group is checked to see if any reindexing may be required to give
    consistent indexing between both datasets. If possible twin operators exist,
    the different indexing options are tested against the reference set, using
    the correlation between datasets as the test.


    Args:
      test_miller_set (cctbx.miller.array): The input miller set to be reindexed.
      reference_miller_set (cctbx.miller.array): The reference miller set.

    Returns:
      cctbx.sgtbx.change_of_basis_op: The change of basis operator which should be
      applied to the test dataset to give consistent indexing with the reference.
    """
    if (
        reference_miller_set.space_group().type().number()
        != test_miller_set.space_group().type().number()
    ):
        raise Sorry(
            """Space groups are not equal. Can only reindex against a
reference dataset if both dataset are in the same spacegroup."""
        )

    twin_ops = twin_laws(miller_array=test_miller_set.eliminate_sys_absent()).operators
    twin_ops = [sgtbx.change_of_basis_op(op.operator.as_xyz()) for op in twin_ops]

    if twin_ops:
        correlations = []
        print(
            "Possible twin operators identified for space group %s:"
            % test_miller_set.space_group().info()
        )
        for op in twin_ops:
            print(op)
        # Loop through twin operators, calculating cc between two datasets
        cc = test_miller_set.correlation(
            reference_miller_set, assert_is_similar_symmetry=False
        )
        correlations.append(cc.coefficient())
        for op in twin_ops:
            reindexed = test_miller_set.change_basis(op)
            cc = reindexed.correlation(
                reference_miller_set, assert_is_similar_symmetry=False
            )
            correlations.append(cc.coefficient())

        # print out table of results and choose best
        header = ["Reindex op", "CC to reference"]
        rows = [["a, b, c (no reindex)", f"{correlations[0]:.5f}"]]
        for i, op in enumerate(twin_ops):
            rows.append([str(op), f"{correlations[i + 1]:.5f}"])
        print(dials.util.tabulate(rows, header))

        best_solution_idx = correlations.index(max(correlations))
        print("\nOutcome of analysis against reference dataset:")
        if best_solution_idx == 0:
            print("No reindexing required \n")
            change_of_basis_op = sgtbx.change_of_basis_op("a,b,c")
        else:
            print(
                "Reindexing required with the twin operator:",
                twin_ops[best_solution_idx - 1].as_hkl(),
                "\n",
            )
            change_of_basis_op = twin_ops[best_solution_idx - 1]
    else:
        print("No twin operators found, no reindexing required \n")
        change_of_basis_op = sgtbx.change_of_basis_op("a,b,c")

    return change_of_basis_op



from __future__ import annotations

from dials.algorithms.integration.bayes.algorithm import IntegrationAlgorithm
from dials_algorithms_integration_bayes_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "BayesianIntegratorDouble",
    "BayesianIntegratorFloat",
    "IntegrationAlgorithm",
    "integrate_by_bayesian_integrator",
)



from __future__ import annotations

import pickle

import procrunner
import pytest

from dxtbx.serialize import load
from scitbx.array_family import flex


def test_model_background(dials_data, tmp_path):
    centroid = dials_data("centroid_test_data", pathlib=True)
    expts = centroid / "experiments.json"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(expts)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmp_path / "trusted_range_patch.expt")

    result = procrunner.run(
        ["dials.model_background", "trusted_range_patch.expt"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    for filename in (
        "background.pickle",
        "mean_0.png",
        "variance_0.png",
        "dispersion_0.png",
        "mask_0.png",
        "min_0.png",
        "max_0.png",
        "model_0.png",
    ):
        assert (tmp_path / filename).is_file()

    with open((tmp_path / "background.pickle"), "rb") as f:
        background = pickle.load(f)

    panel = 0
    data = background.data(panel)
    assert data.all() == (2527, 2463)
    min_max_mean = flex.min_max_mean_double(data.as_1d())
    assert min_max_mean.max == pytest.approx(5.9114028830604095)
    assert min_max_mean.min == 0.0
    assert min_max_mean.mean == pytest.approx(0.5013730161480899)

    # Test integration using the background model, with robust.algorithm=(True|False)
    refls = centroid / "indexed.refl"

    result = procrunner.run(
        [
            "dials.integrate",
            "trusted_range_patch.expt",
            refls,
            "background.algorithm=gmodel",
            "gmodel.robust.algorithm=False",
            "gmodel.model=background.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    result = procrunner.run(
        [
            "dials.integrate",
            "trusted_range_patch.expt",
            refls,
            "background.algorithm=gmodel",
            "gmodel.robust.algorithm=True",
            "gmodel.model=background.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr



from __future__ import annotations

import itertools
import logging
import os
import pathlib

import psutil

import libtbx.easy_mp

import dials.util

logger = logging.getLogger(__name__)


def available_cores() -> int:
    """
    Determine the number of available processor cores.

    There are a number of different methods to get this information, some of
    which may not be available on a specific OS and/or version of Python. So try
    them in order and return the first successful one.
    """

    # https://htcondor.readthedocs.io/en/latest/users-manual/services-for-jobs.html#extra-environment-variables-htcondor-sets-for-jobs
    condor_job_ad = os.environ.get("_CONDOR_JOB_AD")
    if condor_job_ad:
        try:
            classad = dials.util.parse_htcondor_job_classad(pathlib.Path(condor_job_ad))
        except Exception as e:
            logger.error(
                f"Error parsing _CONDOR_JOB_AD {condor_job_ad}: {e}",
                exc_info=True,
            )
        else:
            if classad.cpus_provisioned:
                return classad.cpus_provisioned

    nproc = os.environ.get("NSLOTS", 0)
    try:
        nproc = int(nproc)
        if nproc >= 1:
            return nproc
    except ValueError:
        pass

    try:
        return len(os.sched_getaffinity(0))
    except AttributeError:
        pass

    try:
        return len(psutil.Process().cpu_affinity())
    except AttributeError:
        pass

    nproc = os.cpu_count()
    if nproc is not None:
        return nproc

    nproc = psutil.cpu_count()
    if nproc is not None:
        return nproc

    return 1


class __cluster_function_wrapper:
    """
    A function called by the multi node parallel map. On each cluster node, a
    nested parallel map using the multi processing method will be used.
    """

    def __init__(
        self,
        func,
        nproc=1,
        asynchronous=True,
        preserve_order=True,
    ):
        self.func = func
        self.nproc = nproc
        self.asynchronous = asynchronous
        self.preserve_order = (preserve_order,)

    def __call__(self, iterable):
        return libtbx.easy_mp.parallel_map(
            func=self.func,
            iterable=iterable,
            processes=self.nproc,
            method="multiprocessing",
            asynchronous=self.asynchronous,
            preserve_order=self.preserve_order,
            preserve_exception_message=True,
        )


def _iterable_grouper(iterable, chunk_size):
    """
    Group the iterable into chunks of up to chunk_size items
    """
    args = [iter(iterable)] * chunk_size
    for group in itertools.zip_longest(*args):
        group = tuple(item for item in group if item is not None)
        yield group


class _iterable_wrapper:
    def __init__(self, function):
        self.__function = function

    def __call__(self, iterable):
        return [self.__function(item) for item in iterable]


def multi_node_parallel_map(
    func,
    iterable,
    njobs=1,
    nproc=1,
    cluster_method=None,
    asynchronous=True,
    callback=None,
    preserve_order=True,
):
    """
    A wrapper function to call a function using multiple cluster nodes and with
    multiple processors on each node
    """

    # The function to all on the cluster
    cluster_func = __cluster_function_wrapper(
        func=func,
        nproc=nproc,
        asynchronous=asynchronous,
        preserve_order=preserve_order,
    )

    # Create the cluster iterable
    cluster_iterable = _iterable_grouper(iterable, nproc)

    # Create the cluster callback
    if callback is not None:
        cluster_callback = _iterable_wrapper(callback)
    else:
        cluster_callback = None

    # Do the parallel map on the cluster
    # Call either drmaa or easy_mp to do a parallel map calculation.
    # This function is set up so that in each case we can select
    # the number of cores on a machine
    if cluster_method == "drmaa":
        from dials.util.cluster_map import cluster_map as drmaa_parallel_map

        result = drmaa_parallel_map(
            func=cluster_func,
            iterable=cluster_iterable,
            callback=cluster_callback,
            nslots=nproc,
            njobs=njobs,
            job_category="low",
        )
    else:
        result = libtbx.easy_mp.parallel_map(
            func=cluster_func,
            iterable=cluster_iterable,
            callback=cluster_callback,
            method=cluster_method,
            processes=njobs,
            qsub_command=f"qsub -pe smp {nproc}",
            asynchronous=asynchronous,
            preserve_order=preserve_order,
            preserve_exception_message=True,
        )

    # return result
    return [item for rlist in result for item in rlist]


def batch_multi_node_parallel_map(
    func=None,
    iterable=None,
    nproc=1,
    njobs=1,
    callback=None,
    cluster_method=None,
    chunksize=1,
):
    """
    A function to run jobs in batches in each process
    """
    # Call the batches in parallel
    return multi_node_parallel_map(
        func=_iterable_wrapper(func),
        iterable=_iterable_grouper(iterable, chunksize),
        nproc=nproc,
        njobs=njobs,
        cluster_method=cluster_method,
        callback=_iterable_wrapper(callback),
        preserve_order=True,
    )


if __name__ == "__main__":

    def func(x):
        return x

    iterable = list(range(100))

    multi_node_parallel_map(
        func,
        iterable,
        nproc=4,
        njobs=10,
        cluster_method="multiprocessing",
        callback=print,
    )



from __future__ import annotations


class ThresholdStrategy:
    """
    Base class for spot finder threshold strategies.
    """

    def __init__(self, **kwargs):
        """
        Initialise with key word arguments.
        """
        pass

    def __call__(self, image):
        """
        Threshold the image.
        """
        raise RuntimeError("Overload Me!")


class DispersionThresholdStrategy(ThresholdStrategy):
    """
    A class implementing a 'gain' threshold.
    """

    def __init__(self, **kwargs):
        """
        Set the threshold algorithm up
        """

        # Initialise the base class
        ThresholdStrategy.__init__(self, **kwargs)

        # Get the parameters
        self._kernel_size = kwargs.get("kernel_size", (3, 3))
        self._gain = kwargs.get("gain")
        self._n_sigma_b = kwargs.get("n_sigma_b", 6)
        self._n_sigma_s = kwargs.get("n_sigma_s", 3)
        self._min_count = kwargs.get("min_count", 2)
        self._threshold = kwargs.get("global_threshold", 0)

        # Save the constant gain
        self._gain_map = None

        # Create a buffer
        self.algorithm = {}

    def __call__(self, image, mask):
        """
        Call the thresholding function

        :param image: The image to process
        :param mask: The mask to use
        :return: The thresholded image
        """
        from dials.algorithms.image import threshold
        from dials.array_family import flex

        # Initialise the algorithm
        try:
            algorithm = self.algorithm[image.all()]
        except Exception:
            algorithm = threshold.DispersionThreshold(
                image.all(),
                self._kernel_size,
                self._n_sigma_b,
                self._n_sigma_s,
                self._threshold,
                self._min_count,
            )
            self.algorithm[image.all()] = algorithm

        # Set the gain
        if self._gain is not None:
            assert self._gain > 0
            self._gain_map = flex.double(image.accessor(), self._gain)
            self._gain = None

        # Compute the threshold
        result = flex.bool(flex.grid(image.all()))
        if self._gain_map:
            algorithm(image, mask, self._gain_map, result)
        else:
            algorithm(image, mask, result)

        # Return the result
        return result


class DispersionExtendedThresholdStrategy(ThresholdStrategy):
    """
    A class implementing a 'gain' threshold.
    """

    def __init__(self, **kwargs):
        """
        Set the threshold algorithm up
        """

        # Initialise the base class
        ThresholdStrategy.__init__(self, **kwargs)

        # Get the parameters
        self._kernel_size = kwargs.get("kernel_size", (3, 3))
        self._gain = kwargs.get("gain")
        self._n_sigma_b = kwargs.get("n_sigma_b", 6)
        self._n_sigma_s = kwargs.get("n_sigma_s", 3)
        self._min_count = kwargs.get("min_count", 2)
        self._threshold = kwargs.get("global_threshold", 0)

        # Save the constant gain
        self._gain_map = None

        # Create a buffer
        self.algorithm = {}

    def __call__(self, image, mask):
        """
        Call the thresholding function

        :param image: The image to process
        :param mask: The mask to use
        :return: The thresholded image
        """
        from dials.algorithms.image import threshold
        from dials.array_family import flex

        # Initialise the algorithm
        try:
            algorithm = self.algorithm[image.all()]
        except Exception:
            algorithm = threshold.DispersionExtendedThreshold(
                image.all(),
                self._kernel_size,
                self._n_sigma_b,
                self._n_sigma_s,
                self._threshold,
                self._min_count,
            )
            self.algorithm[image.all()] = algorithm

        # Set the gain
        if self._gain is not None:
            assert self._gain > 0
            self._gain_map = flex.double(image.accessor(), self._gain)
            self._gain = None

        # Compute the threshold
        result = flex.bool(flex.grid(image.all()))
        if self._gain_map:
            algorithm(image, mask, self._gain_map, result)
        else:
            algorithm(image, mask, result)

        # Return the result
        return result



from __future__ import annotations

import numpy as np


def flex_double_as_string(flex_array, n_digits=None):
    if n_digits is not None:
        flex_array = flex_array.round(n_digits)
    return list(flex_array.as_string())


def heatmap_unit_cell_scatter_plots(uc_params, nbins=100, mask_zeros=True):
    """
    Generate heatmap-style unit cell scatter plots.

    If mask_zeros is True, the output will have NaN values where there is no
    data, which will be absent (white) in the resulting plotly plots.
    """

    a, b, c = uc_params[0:3]

    def uc_param_hist2d(p1, p2):
        H, xedges, yedges = np.histogram2d(p1, p2, bins=nbins)
        H = np.rot90(H)
        H = np.flipud(H)
        if mask_zeros:
            Hmasked = np.ma.masked_where(H == 0, H)
            return xedges, yedges, Hmasked
        return xedges, yedges, H

    x1, y1, z1 = uc_param_hist2d(a, b)
    x2, y2, z2 = uc_param_hist2d(b, c)
    x3, y3, z3 = uc_param_hist2d(c, a)
    maxz = max(np.max(z1), np.max(z2), np.max(z3))
    return {
        "data": [
            {
                "x": x1.tolist(),
                "y": y1.tolist(),
                "z": z1.tolist(),
                "type": "heatmap",
                "mode": "markers",
                "name": "a vs. b",
                "colorscale": "Viridis",
                "xaxis": "x",
                "yaxis": "y",
                "zaxis": "z",
                "zmin": 1,
                "zmax": maxz,
            },
            {
                "x": x2.tolist(),
                "y": y2.tolist(),
                "z": z2.tolist(),
                "type": "heatmap",
                "mode": "markers",
                "name": "b vs. c",
                "colorscale": "Viridis",
                "xaxis": "x2",
                "yaxis": "y2",
                "zaxis": "z",
                "zmin": 1,
                "zmax": maxz,
            },
            {
                "x": x3.tolist(),
                "y": y3.tolist(),
                "z": z3.tolist(),
                "type": "heatmap",
                "mode": "markers",
                "name": "c vs. a",
                "colorbar": {
                    "title": "Frequency",
                    "titleside": "right",
                },
                "colorscale": "Viridis",
                "xaxis": "x3",
                "yaxis": "y3",
                "zmin": 1,
                "zmax": maxz,
            },
        ],
        "layout": {
            "title": "Distribution of unit cell parameters",
            "showlegend": False,
            "grid": {"rows": 1, "columns": 3, "subplots": [["xy", "x2y2", "x3y3"]]},
            "xaxis": {"title": "a (Å)"},
            "yaxis": {"title": "b (Å)"},
            "xaxis2": {"title": "b (Å)"},
            "yaxis2": {"title": "c (Å)"},
            "xaxis3": {"title": "c (Å)"},
            "yaxis3": {"title": "a (Å)"},
        },
        "help": """\
The distribution of the unit cell parameters: a vs b, b vs c and c vs a respectively.
""",
    }


def plot_uc_histograms(uc_params, scatter_style="points"):
    a, b, c, al, be, ga = (flex_double_as_string(p, n_digits=4) for p in uc_params)
    if scatter_style == "points":
        uc_scatter_plots = {
            "data": [
                {
                    "x": a,
                    "y": b,
                    "type": "scatter",
                    "mode": "markers",
                    "name": "a vs. b",
                    "xaxis": "x",
                    "yaxis": "y",
                },
                {
                    "x": b,
                    "y": c,
                    "type": "scatter",
                    "mode": "markers",
                    "name": "b vs. c",
                    "xaxis": "x2",
                    "yaxis": "y2",
                },
                {
                    "x": c,
                    "y": a,
                    "type": "scatter",
                    "mode": "markers",
                    "name": "c vs. a",
                    "xaxis": "x3",
                    "yaxis": "y3",
                },
            ],
            "layout": {
                "grid": {"rows": 1, "columns": 3, "pattern": "independent"},
                "title": "Distribution of unit cell parameters",
                "showlegend": False,
                "xaxis": {"title": "a (Å)"},
                "yaxis": {"title": "b (Å)"},
                "xaxis2": {"title": "b (Å)"},
                "yaxis2": {"title": "c (Å)"},
                "xaxis3": {"title": "c (Å)"},
                "yaxis3": {"title": "a (Å)"},
            },
            "help": """\
    The distribution of the unit cell parameters: a vs. b, b vs. c and c vs.a respectively.
    """,
        }
    elif scatter_style == "heatmap":
        uc_scatter_plots = heatmap_unit_cell_scatter_plots(uc_params)
    else:
        raise ValueError("Bad input for scatter_type parameter")

    plots = {
        "uc_scatter": uc_scatter_plots,
        "uc_hist": {
            "data": [
                {
                    "x": a,
                    "type": "histogram",
                    "connectgaps": False,
                    "name": "uc_hist_a",
                    "nbins": "auto",
                    "xaxis": "x",
                    "yaxis": "y",
                },
                {
                    "x": b,
                    "type": "histogram",
                    "connectgaps": False,
                    "name": "uc_hist_b",
                    "nbins": "auto",
                    "xaxis": "x2",
                    "yaxis": "y",
                },
                {
                    "x": c,
                    "type": "histogram",
                    "connectgaps": False,
                    "name": "uc_hist_c",
                    "nbins": "auto",
                    "xaxis": "x3",
                    "yaxis": "y",
                },
            ],
            "layout": {
                "grid": {"rows": 1, "columns": 3, "subplots": [["xy", "x2y", "x3y"]]},
                "title": "Histogram of unit cell parameters",
                "showlegend": False,
                "xaxis": {"title": "a (Å)"},
                "yaxis": {"title": "Frequency"},
                "xaxis2": {"title": "b (Å)"},
                "xaxis3": {"title": "c (Å)"},
            },
            "help": """\
    Histograms of unit cell parameters, a, b and c.
    """,
        },
    }
    uc_angle_hist = {
        "data": [],
        "help": "Histograms of unit cell angles.",
        "layout": {
            "title": "Histogram of unit cell angles",
            "showlegend": False,
            "yaxis": {"title": "Frequency"},
        },
    }
    subplots = []

    n = 0
    # only plot if not symmetry constrained, or only one crystal
    if len(set(al)) > 1 or len(al) == 1:
        n += 1
        uc_angle_hist["data"].append(
            {
                "x": al,
                "type": "histogram",
                "connectgaps": False,
                "name": "uc_hist_alpha",
                "nbins": "auto",
                "xaxis": "x" if n == 1 else f"x{n}",
                "yaxis": "y",
            }
        )
        subplots.append("xy" if n == 1 else f"x{n}y")
        uc_angle_hist["layout"]["xaxis" if n == 1 else f"xaxis{n}"] = {
            "title": "alpha (°)"
        }
    if len(set(be)) > 1 or len(be) == 1:
        n += 1
        uc_angle_hist["data"].append(
            {
                "x": be,
                "type": "histogram",
                "connectgaps": False,
                "name": "uc_hist_beta",
                "nbins": "auto",
                "xaxis": "x" if n == 1 else f"x{n}",
                "yaxis": "y",
            }
        )
        subplots.append("xy" if n == 1 else f"x{n}y")
        uc_angle_hist["layout"]["xaxis" if n == 1 else f"xaxis{n}"] = {
            "title": "beta (°)"
        }
    if len(set(ga)) > 1 or len(ga) == 1:
        n += 1
        uc_angle_hist["data"].append(
            {
                "x": ga,
                "type": "histogram",
                "connectgaps": False,
                "name": "uc_hist_gamma",
                "nbins": "auto",
                "xaxis": "x" if n == 1 else f"x{n}",
                "yaxis": "y",
            }
        )
        subplots.append("xy" if n == 1 else f"x{n}y")
        uc_angle_hist["layout"]["xaxis" if n == 1 else f"xaxis{n}"] = {
            "title": "gamma (°)"
        }
    if n:
        grid = {"rows": 1, "columns": 3, "subplots": [subplots]}
        uc_angle_hist["layout"]["grid"] = grid
        plots["uc_angle_hist"] = uc_angle_hist
    return plots


def scipy_dendrogram_to_plotly_json(ddict, title, xtitle=None, ytitle=None, help=None):
    colors = {
        "b": "rgb(31, 119, 180)",
        "g": "rgb(44, 160, 44)",
        "o": "rgb(255, 127, 14)",
        "r": "rgb(214, 39, 40)",
    }

    dcoord = ddict["dcoord"]
    icoord = ddict["icoord"]
    color_list = ddict["color_list"]
    ivl = ddict["ivl"]

    data = []
    xticktext = []
    xtickvals = []

    for k in range(len(dcoord)):
        x = icoord[k]
        y = dcoord[k]

        if y[0] == 0:
            xtickvals.append(x[0])
        if y[3] == 0:
            xtickvals.append(x[3])

        data.append(
            {
                "x": x,
                "y": y,
                "marker": {"color": colors.get(color_list[k])},
                "mode": "lines",
            }
        )

    xtickvals = sorted(xtickvals)
    xticktext = ivl
    d = {
        "data": data,
        "layout": {
            "barmode": "group",
            "legend": {"x": 100, "y": 0.5, "bordercolor": "transparent"},
            "margin": {"r": 10},
            "showlegend": False,
            "title": title,
            "xaxis": {
                "showline": False,
                "showgrid": False,
                "showticklabels": True,
                "tickangle": 300,
                "title": xtitle,
                "titlefont": {"color": "none"},
                "type": "linear",
                "ticktext": xticktext,
                "tickvals": xtickvals,
                "tickorientation": "vertical",
            },
            "yaxis": {
                "showline": False,
                "showgrid": False,
                "showticklabels": True,
                "tickangle": 0,
                "title": ytitle,
                "type": "linear",
            },
            "hovermode": "closest",
        },
    }
    if help:
        d["help"] = help
    return d



from __future__ import annotations

import logging
import sys
from io import StringIO

from iotbx.phil import parse
from libtbx import Auto

from dials.util import log, show_mail_handle_errors

logger = logging.getLogger("dials.command_line.export")

help_message = """

This program is used to export the results of dials processing in various
formats.

The output formats currently supported are:

MTZ format exports the files as an unmerged mtz file, ready for input to
downstream programs such as Pointless and Aimless. For exporting integrated,
but unscaled data, the required input is a models.expt file and an
integrated.refl file. For exporting scaled data, the required input is a
models.expt file and a scaled.refl file, also passing the option
intensity=scale.

NXS format exports the files as an NXmx file. The required input is a
models.expt file and an integrated.refl file.

MMCIF format exports the files as an mmcif file. The required input is a
models.expt file and an integrated.refl file.

XDS_ASCII format exports intensity data and the experiment metadata in the
same format as used by the output of XDS in the CORRECT step - output can
be scaled with XSCALE.

SADABS format exports intensity data (and geometry by direction cosines)
as an ersatz-SADABS format reverse engineered from the file format used by
EvalCCD for input to SADABS.

MOSFLM format exports the files as an index.mat mosflm-format matrix file and a
mosflm.in file containing basic instructions for input to mosflm. The required
input is an models.expt file.

XDS format exports a models.expt file as XDS.INP and XPARM.XDS files. If a
reflection file is given it will be exported as a SPOT.XDS file.

PETS format exports intensity data and diffraction data in the CIF format
used by PETS. This is primarily intended to produce files suitable for
dynamic diffraction refinement using Jana2020, which requires this format.

Examples::

  # Export to mtz
  dials.export models.expt integrated.refl
  dials.export models.expt integrated.refl mtz.hklout=integrated.mtz
  dials.export models.expt scaled.refl intensity=scale mtz.hklout=scaled.mtz

  # Export to nexus
  dials.export models.expt integrated.refl format=nxs
  dials.export models.expt integrated.refl format=nxs nxs.hklout=integrated.nxs

  # Export to mmcif
  dials.export models.expt integrated.refl format=mmcif
  dials.export models.expt integrated.refl format=mmcif mmcif.hklout=integrated.mmcif

  # Export to mosflm
  dials.export models.expt integrated.refl format=mosflm

  # Export to xds
  dials.export strong.refl format=xds
  dials.export indexed.refl format=xds
  dials.export models.expt format=xds
  dials.export models.expt indexed.refl format=xds
"""

phil_scope = parse(
    """

  format = *mtz sadabs nxs mmcif mosflm xds xds_ascii json shelx pets
    .type = choice
    .help = "The output file format"

  intensity = *auto profile sum scale
    .type = choice(multi=True)
    .help = "Choice of which intensities to export. Allowed combinations:
            scale, profile, sum, profile+sum, sum+profile+scale. Auto will
            default to scale or profile+sum depending on if the data are scaled."


  debug = False
    .type = bool
    .help = "Output additional debugging information"

  mtz {

    combine_partials = True
      .type = bool
      .help = "Combine partials that have the same partial id into one
        reflection, with an updated partiality given by the sum of the
        individual partialities."

    partiality_threshold=0.4
      .type = float
      .help = "All reflections with partiality values above the partiality
        threshold will be retained. This is done after any combination of
        partials if applicable."

    min_isigi = -5
      .type = float
      .help = "Exclude reflections with unfeasible values of I/Sig(I)"

    force_static_model = False
      .type = bool
      .help = "Force program to use static model even if scan varying is present"

    filter_ice_rings = False
      .type = bool
      .help = "Filter reflections at ice ring resolutions"

    d_min = None
      .type = float
      .help = "Filter out reflections with d-spacing below d_min"

    hklout = auto
      .type = path
      .help = "The output MTZ filename, defaults to integrated.mtz or scaled_unmerged.mtz"
              "depending on if the input data are scaled."

    crystal_name = XTAL
      .type = str
      .help = "The name of the crystal, for the mtz file metadata"

    project_name = DIALS
      .type = str
      .help = "The project name for the mtz file metadata"

    best_unit_cell = None
    .type = unit_cell
    .help = "Best unit cell value, to use when performing resolution cutting,"
            "and as the overall unit cell in the exported mtz. If None, the median"
            "cell will be used."

  }

  sadabs {

    hklout = integrated.sad
      .type = path
      .help = "The output raw sadabs file"
    run = 1
      .type = int
      .help = "Batch number / run number for output file"
    predict = False
      .type = bool
      .help = "Compute centroids with static model, not observations"

  }

  xds_ascii {

    hklout = DIALS.HKL
      .type = path
      .help = "The output raw hkl file"

  }

  nxs {

    hklout = integrated.nxs
      .type = path
      .help = "The output Nexus file"

    instrument_name = Unknown
      .type = str
      .help = "Name of the instrument/beamline"

    instrument_short_name = Unknown
      .type = str
      .help = "Short name for instrument/beamline, perhaps the acronym"

    source_name = Unknown
      .type = str
      .help = "Name of the source/facility"

    source_short_name = Unknown
      .type = str
      .help = "Short name for source, perhaps the acronym"

  }

  mmcif {

    hklout = auto
      .type = path
      .help = "The output CIF file, defaults to integrated.cif or scaled_unmerged.cif
        depending on if the data are scaled."

    compress = gz bz2 xz
      .type = choice
      .help = "Choose compression format (also appended to the file name)"

    pdb_version = *v5 v5_next
      .type = choice
      .help = "This controls which pdb mmcif dictionary version the output"
              "mmcif file should comply with. v5_next adds support for"
              "recording unmerged data as well as additional scan metadata"
              "and statistics, however writing can be slow for large datasets."
  }

  mosflm {

    directory = mosflm
      .type = path
      .help = "The output directory for mosflm output"

  }

  xds {

    directory = xds
      .type = path
      .help = "The output directory for xds output"

  }

  json {

    filename = rlp.json
      .type = path
    compact = True
      .type = bool
    n_digits = 6
      .type = int(value_min=1)
      .help = "Number of decimal places to be used for representing the"
              "reciprocal lattice points."

  }

  shelx {

    hklout = dials.hkl
      .type = path
      .help = "The output hkl file"
    ins = dials.ins
      .type = path
      .help = "The output ins file"
    scale = True
      .type = bool
      .help = "Scale reflections to maximise output precision in SHELX 8.2f format"
    scale_range = -9999.0, 9999.0
      .type = floats(size=2, value_min=-999999., value_max=9999999.)
      .help = "minimum or maximum intensity value after scaling."

  }

  pets {

    filename_prefix = dials_dyn
      .type = str
      .help = "The prefix for output files, where the default will produce"
              "dials_dyn.cif_pets"
    id = None
      .type = int
      .help = "The experiment ID to export from a multi-experiment list"

    partiality_cutoff = 0.99
      .type = float
      .help = "Cutoff for determining which reflections are deemed to be fully"
              "recorded"

    flag_filtering = False
      .type = bool
      .help = "If true, keep only the reflections where the relevant `integrated`"
              "flag is set (either `integrated_sum` or `integrated_prf`). This"
              "seems to be quite restrictive compared to"
              "PETS, so is not set by default."

    virtual_frame {
      excitation_error_cutoff = 0.04
        .type = float
        .help = "Excitation error cutoff determining which reflections are"
                "included in virtual frames"
      n_merged = 1
        .type = int
        .help = "Number of frames to merge in a virtual frame"
      step = 1
        .type = int
        .help = "Step between frames"

    }
  }

  output {
    log = dials.export.log
      .type = path
      .help = "The log filename"
  }
"""
)


def _check_input(experiments, reflections, params=None, check_intensities=False):
    if not experiments:
        raise ValueError("export requires an experiment list")
    if len(reflections) != 1:
        raise ValueError("export requires 1 reflection table")
    if check_intensities:
        if not params:
            raise ValueError("No parameters given")
        if "profile" not in params.intensity and "sum" not in params.intensity:
            raise ValueError(
                "Only intensity options containing sum or profile can be exported in this format"
            )
        if (
            "intensity.sum.value" not in reflections[0]
            and "intensity.prf.value" not in reflections[0]
        ):
            raise ValueError(
                "Unable to find 'intensity.sum.value' or 'intensity.prf.value' columns in reflection table."
            )


def export_mtz(params, experiments, reflections):
    """
    Export reflections in MTZ format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections)

    from dials.util.export_mtz import export_mtz

    # Handle case where user has passed data before integration
    if (
        "intensity.sum.value" not in reflections[0]
        and "intensity.prf.value" not in reflections[0]
    ):
        raise ValueError(
            "Error: No intensity data in reflections; cannot export un-integrated data to MTZ"
        )

    reflection_table = reflections[0]
    filename = params.mtz.hklout
    # if mtz filename is auto, then choose scaled.mtz or integrated.mtz
    if filename in (None, Auto, "auto"):
        if ("intensity.scale.value" in reflection_table) and (
            "intensity.scale.variance" in reflection_table
        ):
            filename = "scaled.mtz"
            logger.info("Data appears to be scaled, setting mtz.hklout = 'scaled.mtz'")
        else:
            filename = "integrated.mtz"
            logger.info(
                "Data appears to be unscaled, setting mtz.hklout = 'integrated.mtz'"
            )

    m = export_mtz(
        reflection_table,
        experiments,
        intensity_choice=params.intensity,
        filename=filename,
        best_unit_cell=params.mtz.best_unit_cell,
        partiality_threshold=params.mtz.partiality_threshold,
        combine_partials=params.mtz.combine_partials,
        min_isigi=params.mtz.min_isigi,
        filter_ice_rings=params.mtz.filter_ice_rings,
        d_min=params.mtz.d_min,
        force_static_model=params.mtz.force_static_model,
        crystal_name=params.mtz.crystal_name,
        project_name=params.mtz.project_name,
    )

    summary = StringIO()
    m.show_summary(out=summary)
    logger.info("")
    logger.info(summary.getvalue())


def export_sadabs(params, experiments, reflections):
    """
    Export data in HKL format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections, params=params, check_intensities=True)

    from dials.util.export_sadabs import export_sadabs

    export_sadabs(reflections[0], experiments, params)


def export_xdsascii(params, experiments, reflections):
    """
    Export data in XDS_ASCII format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections, params=params, check_intensities=True)

    from dials.util.export_xds_ascii import export_xds_ascii

    export_xds_ascii(reflections[0], experiments, params)


def export_nexus(params, experiments, reflections):
    """
    Export data in Nexus format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections)

    from dials.util.nexus import dump

    dump(experiments, reflections[0], params.nxs)


def export_mmcif(params, experiments, reflections):
    """
    Export data in CIF format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections)

    from dials.util.export_mmcif import MMCIFOutputFile

    outfile = MMCIFOutputFile(params)
    outfile.write(experiments, reflections[0])


def export_mosflm(params, experiments, reflections):
    """
    Export stuff in mosflm format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, [None])

    if reflections:
        sys.exit("Mosflm exporter does not need a reflection table")

    from dials.util.mosflm import dump

    dump(experiments, params.mosflm.directory)


def export_xds(params, experiments, reflections):
    """
    Export stuff in xds format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    # Check the input
    if len(reflections) > 1:
        raise ValueError("XDS exporter requires 0 or 1 reflection table")

    if reflections:
        reflections = reflections[0]

    from dials.util.xds import dump

    dump(experiments, reflections, params.xds.directory)


def export_json(params, experiments, reflections):
    """
    Export reflections in json format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    # Check the input
    _check_input(experiments, [None])
    if not reflections:
        raise ValueError("json exporter requires a reflection table")

    from scitbx.array_family import flex

    from dials.util import export_json

    imagesets = [expt.imageset for expt in experiments]

    if len(reflections) != len(imagesets):
        raise ValueError(
            "Mismatch between %d reflections lists and %d imagesets"
            % (len(reflections), len(imagesets))
        )
    selected_reflections = None
    for i, refl in enumerate(reflections):
        refl["imageset_id"] = flex.size_t(refl.size(), i)
        if selected_reflections is None:
            selected_reflections = refl
        else:
            selected_reflections.extend(refl)

    exporter = export_json.ReciprocalLatticeJson(experiments, selected_reflections)
    exporter.as_json(
        filename=params.json.filename,
        compact=params.json.compact,
        n_digits=params.json.n_digits,
        experiments=experiments,
    )


def export_shelx(params, experiments, reflections):
    """
    Export data in SHELX HKL format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    _check_input(experiments, reflections, params=params)

    # check for a single intensity choice
    if len(params.intensity) > 1:
        raise ValueError(
            "Only 1 intensity option can be exported in this format, please choose a single intensity option e.g. intensity=profile"
        )

    from dials.util.export_shelx import export_shelx

    export_shelx(reflections[0], experiments, params)


def export_pets(params, experiments, reflections):
    """
    Export reflections in PETS CIF format

    :param params: The phil parameters
    :param experiments: The experiment list
    :param reflections: The reflection tables
    """

    # Check the input
    _check_input(experiments, reflections, check_intensities=False)

    from dials.util.export_pets import PETSOutput

    pets_output = PETSOutput(experiments, reflections, params)
    pets_output.write_dyn_cif_pets()

    return


@show_mail_handle_errors()
def run(args=None):
    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )
    from dials.util.version import dials_version

    usage = "dials.export models.expt reflections.refl [options]"

    # Create the option parser
    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        phil=phil_scope,
        epilog=help_message,
    )

    # Get the parameters
    params, options = parser.parse_args(args, show_diff_phil=False)

    # Configure the logging
    log.config(logfile=params.output.log)

    # Print the version number
    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    if not params.input.experiments and not params.input.reflections:
        parser.print_help()
        sys.exit()

    # Get the experiments and reflections
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    # do auto interpreting of intensity choice:
    # note that this may still fail certain checks further down the processing,
    # but these are the defaults to try
    if params.intensity in ([None], [Auto], ["auto"], Auto) and reflections:
        if ("intensity.scale.value" in reflections[0]) and (
            "intensity.scale.variance" in reflections[0]
        ):
            params.intensity = ["scale"]
            logger.info("Data appears to be scaled, setting intensity = scale")
        else:
            params.intensity = []
            if "intensity.sum.value" in reflections[0]:
                params.intensity.append("sum")
            if "intensity.prf.value" in reflections[0]:
                params.intensity.append("profile")
            logger.info(
                "Data appears to be unscaled, setting intensity = "
                + "+".join(params.intensity)
            )

    # Choose the exporter
    exporter = {
        "mtz": export_mtz,
        "sadabs": export_sadabs,
        "xds_ascii": export_xdsascii,
        "nxs": export_nexus,
        "mmcif": export_mmcif,
        "mosflm": export_mosflm,
        "xds": export_xds,
        "json": export_json,
        "shelx": export_shelx,
        "pets": export_pets,
    }.get(params.format)
    if not exporter:
        sys.exit(f"Unknown format: {params.format}")

    # Export the data
    try:
        exporter(params, experiments, reflections)
    except Exception as e:
        logger.error(f"Error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    run()



"""Basis vector search strategies."""


from __future__ import annotations


class Strategy:
    """A base class for basis vector search strategies."""

    phil_help = None

    phil_scope = None

    def __init__(self, max_cell, params=None, *args, **kwargs):
        """Construct the strategy.

        Args:
            max_cell (float): An estimate of the maximum cell dimension of the primitive
                cell.
        """
        self._max_cell = max_cell
        self._params = params
        if self._params is None and self.phil_scope is not None:
            self._params = self.phil_scope.extract()

    def find_basis_vectors(self, reciprocal_lattice_vectors):
        """Find a list of likely basis vectors.

        Args:
            reciprocal_lattice_vectors (scitbx.array_family.flex.vec3_double):
                The list of reciprocal lattice vectors to search for periodicity.

        Returns:
            A tuple containing the list of basis vectors and a flex.bool array
            identifying which reflections were used in indexing.
        """
        raise NotImplementedError()



from __future__ import annotations

from libtbx import easy_run


def test_show_extensions():
    # Call dials.merge_reflection_lists
    easy_run.fully_buffered(["dev.dials.show_extensions"]).raise_if_errors()



from __future__ import annotations

from scitbx import matrix


def is_approximate_integer_multiple(
    vec_a, vec_b, relative_length_tolerance=0.2, angular_tolerance=5.0
):
    length_a = vec_a.length()
    length_b = vec_b.length()
    # assert length_b >= length_a
    if length_a > length_b:
        vec_a, vec_b = vec_b, vec_a
        length_a, length_b = length_b, length_a
    angle = vec_a.angle(vec_b, deg=True)
    if angle < angular_tolerance or abs(180 - angle) < angular_tolerance:
        n = length_b / length_a
        if abs(round(n) - n) < relative_length_tolerance:
            return True
    return False


class vector_group:
    def __init__(self):
        self.vectors = []
        self.weights = []
        self._mean = None

    def append(self, vector, weight):
        self.vectors.append(vector)
        self.weights.append(weight)
        self._mean = None

    @property
    def mean(self):
        if self._mean is None:
            self._mean = self._compute_mean()
        return self._mean

    def _compute_mean(self):
        sum_x = 0
        sum_y = 0
        sum_z = 0
        for v in self.vectors:
            sum_x += v.elems[0]
            sum_y += v.elems[1]
            sum_z += v.elems[2]
        return matrix.col((sum_x, sum_y, sum_z)) / len(self.vectors)


def group_vectors(
    vectors,
    weights,
    relative_length_tolerance=0.1,
    angular_tolerance=5,
    max_groups=None,
):
    # loop over vectors and sort into groups similar to further down
    # group similar angle and lengths, also catch integer multiples of vectors

    vector_groups = []

    for v, weight in zip(vectors, weights):
        if max_groups is not None and len(vector_groups) == max_groups:
            break
        if not isinstance(v, matrix.col):
            v = matrix.col(v)
        length = v.length()
        matched_group = False
        for group in vector_groups:
            mean_v = group.mean
            mean_v_length = mean_v.length()
            if (
                abs(mean_v_length - length) / max(mean_v_length, length)
                < relative_length_tolerance
            ):
                angle = mean_v.angle(v, deg=True)
                if angle < angular_tolerance:
                    group.append(v, weight)
                    matched_group = True
                    break
                elif abs(180 - angle) < angular_tolerance:
                    group.append(-v, weight)
                    matched_group = True
                    break
        if not matched_group:
            group = vector_group()
            group.append(v, weight)
            vector_groups.append(group)

    return vector_groups



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1


from __future__ import annotations

import json
import sys

import libtbx
import libtbx.phil
from scitbx.array_family import flex

import dials.util
from dials.util import Sorry

help_message = """
Generate a 1d or 2d goniometer detector shadow plot for a given experiment list.

Examples::

  dials.shadow_plot models.expt

  dials.shadow_plot models.expt mode=2d
"""

phil_scope = libtbx.phil.parse(
    """
oscillation_range = None
  .type = floats(size=2)
step_size = auto
  .type = float(value_min=0)
y_max = None
  .type = float(value_min=0, value_max=100)
mode = *1d 2d
  .type = choice
output {
  plot = scan_shadow_plot.png
    .type = path
  json = None
    .type = path
  size_inches = None
    .type = floats(value_min=0, size=2)
}
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import ArgumentParser, flatten_experiments

    usage = "dials.shadow_plot [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=True,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        sys.exit(0)

    assert len(experiments) == 1
    imagesets = experiments.imagesets()

    imageset = imagesets[0]
    goniometer = imageset.get_goniometer()
    detector = imageset.get_detector()
    scan = imageset.get_scan()
    masker = imageset.masker()
    if masker is None:
        raise Sorry("Goniometer model does not support shadowing.")
    angles = goniometer.get_angles()
    names = goniometer.get_names()
    scan_axis = goniometer.get_scan_axis()
    phi = angles[0]

    if params.step_size is libtbx.Auto:
        if params.mode == "1d":
            step = scan.get_oscillation()[1]
        else:
            step = 10
    else:
        step = params.step_size

    if params.mode == "1d":
        if params.oscillation_range is not None:
            start, end = params.oscillation_range
        else:
            start, end = scan.get_oscillation_range()

        scan_points = flex.double(libtbx.utils.frange(start, end, step=step))
        n_px_shadowed = flex.double(scan_points.size(), 0)
        n_px_tot = flex.double(scan_points.size(), 0)

        assert len(angles) == 3
        for i, scan_angle in enumerate(scan_points):
            shadow = masker.project_extrema(detector, scan_angle)
            for p_id in range(len(detector)):
                px_x, px_y = detector[p_id].get_image_size()
                n_px_tot[i] += px_x * px_y
                if shadow[p_id].size() < 4:
                    continue
                n_px_shadowed[i] += polygon_area(shadow[p_id])

    else:
        kappa_values = flex.double(libtbx.utils.frange(0, 360, step=step))
        omega_values = flex.double(libtbx.utils.frange(0, 360, step=step))
        grid = flex.grid(kappa_values.size(), omega_values.size())
        n_px_shadowed = flex.double(grid, 0)
        n_px_tot = flex.double(grid, 0)

        assert len(angles) == 3
        for i, kappa in enumerate(kappa_values):
            for j, omega in enumerate(omega_values):
                masker.set_goniometer_angles((phi, kappa, omega))
                masker.extrema_at_scan_angle(omega)
                shadow = masker.project_extrema(detector, omega)
                for p_id in range(len(detector)):
                    px_x, px_y = detector[p_id].get_image_size()
                    n_px_tot[i, j] += px_x * px_y
                    if shadow[p_id].size() < 4:
                        continue
                    n_px_shadowed[i, j] += polygon_area(shadow[p_id])

    fraction_shadowed = n_px_shadowed / n_px_tot

    if params.output.json is not None:
        if params.mode == "2d":
            raise Sorry("json output not supported for mode=2d")

        print(f"Writing json output to {params.output.json}")
        d = {
            "scan_points": list(scan_points),
            "fraction_shadowed": list(fraction_shadowed),
        }
        with open(params.output.json, "w") as f:
            json.dump(d, f)

    if params.output.plot is not None:
        import matplotlib

        matplotlib.use("Agg")
        from matplotlib import pyplot as plt

        plt.style.use("ggplot")

        if params.mode == "1d":
            plt.plot(
                scan_points.as_numpy_array(), fraction_shadowed.as_numpy_array() * 100
            )
            plt.xlabel(f"{names[scan_axis]} angle (degrees)")
            plt.ylabel("Shadowed area (%)")
            if params.y_max is not None:
                plt.ylim(0, params.y_max)
            else:
                plt.ylim(0, plt.ylim()[1])
        else:
            fig = plt.imshow(
                fraction_shadowed.as_numpy_array() * 100, interpolation="bicubic"
            )
            plt.xlabel(f"{names[2]} angle (degrees)")
            plt.ylabel(f"{names[1]} angle (degrees)")
            plt.xlim(0, 360 / step - 0.5)
            plt.ylim(0, 360 / step - 0.5)

            ticks = (0, 50, 100, 150, 200, 250, 300, 350)
            fig.axes.xaxis.set_major_locator(
                matplotlib.ticker.FixedLocator([k / step for k in ticks])
            )
            fig.axes.yaxis.set_major_locator(
                matplotlib.ticker.FixedLocator([k / step for k in ticks])
            )
            fig.axes.set_xticklabels([f"{k:.0f}" for k in ticks])
            fig.axes.set_yticklabels([f"{k:.0f}" for k in ticks])
            cbar = plt.colorbar()
            cbar.set_label("Shadowed area (%)")

        if params.output.size_inches is not None:
            fig = plt.gcf()
            fig.set_size_inches(params.output.size_inches)
        plt.tight_layout()
        print(f"Saving plot to {params.output.plot}")
        plt.savefig(params.output.plot)


def polygon_area(points):
    # http://mathworld.wolfram.com/PolygonArea.html
    x0, y0 = points.parts()
    x1 = x0[1:]
    x1.append(x0[0])
    y1 = y0[1:]
    y1.append(y0[0])

    return 0.5 * abs(flex.sum(x0 * y1 - x1 * y0))


if __name__ == "__main__":
    run()



"""
Tests for dials.util.multi_dataset_handling functions
"""

from __future__ import annotations

import pytest

from dxtbx.model import Experiment, ExperimentList
from dxtbx.serialize import load

from dials.array_family import flex
from dials.util.multi_dataset_handling import (
    assign_unique_identifiers,
    parse_multiple_datasets,
    renumber_table_id_columns,
    select_datasets_on_ids,
    sort_tables_to_experiments_order,
    update_imageset_ids,
)

from . import mock_reflection_file_object, mock_two_reflection_file_object


@pytest.fixture
def experiments():
    """Make a list of three empty experiments"""
    experiments = ExperimentList()
    experiments.append(Experiment())
    experiments.append(Experiment())
    experiments.append(Experiment())
    return experiments


@pytest.fixture
def experiments_024(experiments):
    experiments[0].identifier = "0"
    experiments[1].identifier = "2"
    experiments[2].identifier = "4"
    return experiments


@pytest.fixture
def reflections():
    """Make a list of three reflection tables"""
    rt1 = flex.reflection_table()
    rt1["id"] = flex.int([0, 0, 0])
    rt2 = flex.reflection_table()
    rt2["id"] = flex.int([1, 1])
    rt3 = flex.reflection_table()
    rt3["id"] = flex.int([4, 4])
    reflections = [rt1, rt2, rt3]
    return reflections


@pytest.fixture
def reflections_024(reflections):
    reflections[0].experiment_identifiers()[0] = "0"
    reflections[1].experiment_identifiers()[1] = "2"
    reflections[2].experiment_identifiers()[4] = "4"
    return reflections


def test_select_specific_datasets_using_id(experiments_024, reflections_024):
    use_datasets = [0, 1]
    experiments, refl = select_datasets_on_ids(
        experiments_024, reflections_024, use_datasets=use_datasets
    )
    assert len(experiments) == 2
    assert len(refl) == 2
    assert list(experiments.identifiers()) == ["0", "2"]


def test_exclude_specific_datasets_using_id(experiments_024, reflections_024):
    experiments, refl = select_datasets_on_ids(
        experiments_024, reflections_024, exclude_datasets=["0"]
    )
    assert len(refl) == 2
    assert list(experiments.identifiers()) == ["2", "4"]
    assert len(experiments) == 2


def test_raise_exception_when_selecting_and_excluding_datasets_at_same_time(
    experiments_024, reflections_024
):
    with pytest.raises(ValueError):
        experiments, refl = select_datasets_on_ids(
            experiments_024,
            reflections_024,
            use_datasets=["2", "4"],
            exclude_datasets=["0"],
        )


def test_raise_exception_when_excluding_non_existing_dataset(
    experiments_024, reflections_024
):
    with pytest.raises(ValueError):
        experiments, refl = select_datasets_on_ids(
            experiments_024, reflections_024, exclude_datasets=["3"]
        )


def test_selecting_everything_is_identity_function(experiments_024, reflections_024):
    exp, refl = select_datasets_on_ids(experiments_024, reflections_024)
    assert exp is experiments_024
    assert refl is reflections_024


def test_raise_exception_when_not_all_identifiers_set(experiments, reflections_024):
    experiments[0].identifier = "0"
    experiments[1].identifier = "2"
    with pytest.raises(ValueError):
        exp, refl = select_datasets_on_ids(
            experiments, reflections_024, use_datasets=["2"]
        )


def test_raise_exception_when_selecting_non_existing_dataset(
    experiments_024, reflections_024
):
    with pytest.raises(ValueError):
        exp, refl = select_datasets_on_ids(
            experiments_024, reflections_024, use_datasets=["3"]
        )


def test_correct_handling_with_multi_dataset_table(experiments_024):
    reflections = flex.reflection_table()
    reflections["id"] = flex.int([0, 1, 2])
    reflections.experiment_identifiers()[0] = "0"
    reflections.experiment_identifiers()[1] = "2"
    reflections.experiment_identifiers()[2] = "4"
    exp, refl = select_datasets_on_ids(
        experiments_024,
        [reflections],
        exclude_datasets=["1"],  # i.e. the second dataset
    )
    assert list(refl[0].experiment_identifiers().values()) == ["0", "4"]
    assert list(refl[0]["id"]) == [0, 2]


def test_assignment_of_unique_identifiers_when_refl_table_ids_are_present(
    experiments, reflections
):
    assert list(experiments.identifiers()) == ["", "", ""]
    expts, rts = assign_unique_identifiers(experiments, reflections)
    for i, (expt, refl) in enumerate(zip(expts, rts)):
        assert set(refl["id"]) == {i}
        assert expt.identifier != ""
        assert refl.experiment_identifiers()[i] == expt.identifier


def test_assign_identifiers_where_none_are_set_but_refl_table_ids_have_duplicates(
    experiments, reflections
):
    reflections[2]["id"] = flex.int([0, 0])
    expts, rts = assign_unique_identifiers(experiments, reflections)
    for i, (expt, refl) in enumerate(zip(expts, rts)):
        assert set(refl["id"]) == {i}
        assert expt.identifier != ""
        assert refl.experiment_identifiers()[i] == expt.identifier


def test_raise_exception_when_existing_identifiers_are_inconsistent(
    experiments_024, reflections
):
    reflections[1].experiment_identifiers()[0] = "5"
    # should raise an assertion error for inconsistent identifiers
    with pytest.raises(ValueError):
        _, __ = assign_unique_identifiers(experiments_024, reflections)


def test_cases_where_all_set_whether_reflection_table_is_split_or_not(
    experiments_024, reflections_024
):
    # should pass experiments back if identifiers all already set
    exp, rts = assign_unique_identifiers(experiments_024, reflections_024)
    expected_identifiers = ["0", "2", "4"]
    # Check that identifiers are set in experiments and reflection table.
    assert exp is experiments_024
    assert list(exp.identifiers()) == expected_identifiers
    expected_ids = [0, 1, 4]
    for i, refl in enumerate(rts):
        id_ = refl["id"][0]
        assert refl.experiment_identifiers()[id_] == expected_identifiers[i]
        assert set(refl["id"]) == {id_}
        assert id_ == expected_ids[i]


def test_raise_exception_if_unequal_experiments_and_reflections(experiments_024):
    reflections_multi = [flex.reflection_table()]
    reflections_multi[0]["id"] = flex.int([0, 1, 4])
    reflections_multi[0].experiment_identifiers()[0] = "0"
    reflections_multi[0].experiment_identifiers()[1] = "2"
    reflections_multi[0].experiment_identifiers()[4] = "4"
    del reflections_multi[0].experiment_identifiers()[4]
    del reflections_multi[0]["id"][2]
    with pytest.raises(ValueError):
        _, __ = assign_unique_identifiers(experiments_024, reflections_multi)


def test_assigned_identifiers_are_kept_when_assigning_rest(experiments, reflections):
    # Now test that if some are set, these are maintained and unique ids are
    # set for the rest
    experiments[0].identifier = "1"
    reflections[0].experiment_identifiers()[0] = "1"
    expts, rts = assign_unique_identifiers(experiments, reflections)
    assert expts.identifiers()[0] == "1"
    for i, (expt, refl) in enumerate(zip(expts, rts)):
        assert set(refl["id"]) == {i}
        assert expt.identifier != ""
        assert refl.experiment_identifiers()[i] == expt.identifier


def test_assigning_specified_identifiers(experiments, reflections):
    reflections[0].experiment_identifiers()[0] = "5"
    reflections[1].experiment_identifiers()[1] = "6"
    reflections[1].experiment_identifiers()[4] = "7"
    input_identifiers = ["0", "1", "10"]
    exp, rts = assign_unique_identifiers(experiments, reflections, input_identifiers)
    assert list(exp.identifiers()) == input_identifiers
    assert rts[0].experiment_identifiers()[0] == "0"
    assert rts[0]["id"][0] == 0
    assert rts[1].experiment_identifiers()[1] == "1"
    assert rts[1]["id"][0] == 1
    assert rts[2].experiment_identifiers()[2] == "10"
    assert rts[2]["id"][0] == 2

    # Test raises ValueError when wrong number of identifiers given
    with pytest.raises(ValueError):
        exp, rts = assign_unique_identifiers(experiments, reflections, ["0", "1"])


def test_parse_multiple_datasets():
    """Test the namesake function. This expects a list of reflection tables, and
    selects on the column named 'id'."""
    rt1 = flex.reflection_table()
    rt1["id"] = flex.int([0, 0, 0])
    rt1.experiment_identifiers()[0] = "0"
    rt2 = flex.reflection_table()
    rt2["id"] = flex.int([2, 2, 4, 4])
    rt2.experiment_identifiers()[2] = "2"
    rt2.experiment_identifiers()[4] = "4"
    single_tables = parse_multiple_datasets([rt2])
    assert len(single_tables) == 2
    assert list(set(single_tables[0]["id"])) == [2]
    assert list(set(single_tables[1]["id"])) == [4]
    single_tables = parse_multiple_datasets([rt1, rt2])
    assert list(set(single_tables[0]["id"])) == [0]
    assert list(set(single_tables[1]["id"])) == [2]
    assert list(set(single_tables[2]["id"])) == [4]
    assert len(single_tables) == 3
    single_tables = parse_multiple_datasets([rt1])
    assert len(single_tables) == 1
    assert list(set(single_tables[0]["id"])) == [0]

    # if a duplicate id is given, then this should be detected and new ids
    # determined for all datasets.
    rt3 = flex.reflection_table()
    rt3["id"] = flex.int([2, 2])
    rt3.experiment_identifiers()[2] = "5"
    single_tables = parse_multiple_datasets([rt1, rt2, rt3])
    assert len(single_tables) == 4
    assert list(set(single_tables[0]["id"])) == [0]
    assert single_tables[0].experiment_identifiers()[0] == "0"
    assert list(set(single_tables[1]["id"])) == [1]
    assert single_tables[1].experiment_identifiers()[1] == "2"
    assert list(set(single_tables[2]["id"])) == [2]
    assert single_tables[2].experiment_identifiers()[2] == "4"
    assert list(set(single_tables[3]["id"])) == [3]
    assert single_tables[3].experiment_identifiers()[3] == "5"


def test_sort_tables_to_experiments_order_multi_dataset_files():
    """Test reflection table sorting when a table contains multiple datasets."""
    # Reflection tables in the wrong order
    reflection_tables = [
        mock_two_reflection_file_object(ids=[1, 2]).data,
        mock_reflection_file_object(id_=0).data,
    ]
    experiments = ExperimentList()
    experiments.append(Experiment(identifier=str(0)))
    experiments.append(Experiment(identifier=str(1)))
    experiments.append(Experiment(identifier=str(2)))

    refls = sort_tables_to_experiments_order(reflection_tables, experiments)

    # Check that reflection tables are rearranged
    assert refls[0] is reflection_tables[1]
    assert refls[1] is reflection_tables[0]
    assert list(refls[0].experiment_identifiers().values()) == ["0"]
    assert list(refls[1].experiment_identifiers().values()) == ["1", "2"]


def test_renumber_table_id_columns():
    """Test the correct handling of duplicate table id values.
    Note that this function does not have the ability to update the
    experiment string identifier, only ensure that the table id values
    do not clash.
    """
    # Test the case of two single reflection tables.
    rs = [
        mock_reflection_file_object(id_=0).data,
        mock_reflection_file_object(id_=0).data,
    ]
    rs = renumber_table_id_columns(rs)
    assert list(rs[0]["id"]) == [-1, 0, 0]
    assert list(rs[0].experiment_identifiers().keys()) == [0]
    assert list(rs[0].experiment_identifiers().values()) == ["0"]
    assert list(rs[1]["id"]) == [-1, 1, 1]
    assert list(rs[1].experiment_identifiers().keys()) == [1]
    assert list(rs[1].experiment_identifiers().values()) == ["0"]

    # Now test the case where one reflection table contains two experiments
    rs = [
        mock_two_reflection_file_object().data,
        mock_reflection_file_object(id_=0).data,
    ]
    rs = renumber_table_id_columns(rs)
    assert list(rs[0]["id"]) == [-1, 0, 0, 1, 1]
    assert list(rs[0].experiment_identifiers().keys()) == [0, 1]
    assert list(rs[0].experiment_identifiers().values()) == ["0", "2"]
    assert list(rs[1]["id"]) == [-1, 2, 2]
    assert list(rs[1].experiment_identifiers().keys()) == [2]
    assert list(rs[1].experiment_identifiers().values()) == ["0"]

    rs = [
        mock_reflection_file_object(id_=0).data,
        mock_two_reflection_file_object(ids=[1, 2]).data,
    ]
    rs = renumber_table_id_columns(rs)
    assert list(rs[0]["id"]) == [-1, 0, 0]
    assert list(rs[0].experiment_identifiers().keys()) == [0]
    assert list(rs[0].experiment_identifiers().values()) == ["0"]
    assert list(rs[1]["id"]) == [-1, 1, 1, 2, 2]
    assert list(rs[1].experiment_identifiers().keys()) == [1, 2]
    assert list(rs[1].experiment_identifiers().values()) == ["1", "2"]

    rs = [
        mock_two_reflection_file_object(ids=[1, 2]).data,
        mock_reflection_file_object(id_=0).data,
    ]
    rs = renumber_table_id_columns(rs)
    assert list(rs[0]["id"]) == [-1, 0, 0, 1, 1]
    assert list(rs[0].experiment_identifiers().keys()) == [0, 1]
    assert list(rs[0].experiment_identifiers().values()) == ["1", "2"]
    assert list(rs[1]["id"]) == [-1, 2, 2]
    assert list(rs[1].experiment_identifiers().keys()) == [2]
    assert list(rs[1].experiment_identifiers().values()) == ["0"]


def test_sort_tables_to_experiments_order_single_dataset_files():
    """Test reflection table sorting when tables contain a single dataset."""
    # Reflection tables in the wrong order
    reflection_tables = [
        mock_reflection_file_object(id_=1).data,
        mock_reflection_file_object(id_=0).data,
    ]
    experiments = ExperimentList()
    experiments.append(Experiment(identifier=str(0)))
    experiments.append(Experiment(identifier=str(1)))
    refls = sort_tables_to_experiments_order(reflection_tables, experiments)

    # Check that reflection tables are rearranged
    assert refls[0] is reflection_tables[1]
    assert refls[1] is reflection_tables[0]
    assert list(refls[0].experiment_identifiers().values()) == ["0"]
    assert list(refls[1].experiment_identifiers().values()) == ["1"]

    # Reflection tables in correct order
    reflection_tables = [
        mock_reflection_file_object(id_=0).data,
        mock_reflection_file_object(id_=1).data,
    ]
    experiments = ExperimentList()
    experiments.append(Experiment(identifier=str(0)))
    experiments.append(Experiment(identifier=str(1)))
    refls = sort_tables_to_experiments_order(reflection_tables, experiments)

    # Check that nothing has been changed
    assert refls[0] is reflection_tables[0]
    assert refls[1] is reflection_tables[1]
    assert list(refls[0].experiment_identifiers().values()) == ["0"]
    assert list(refls[1].experiment_identifiers().values()) == ["1"]


def test_update_imageset_ids(dials_data):
    expts = ExperimentList()
    refls = []
    for i in [1, 2, 3, 4, 5, 7, 8, 10]:
        refls.append(
            flex.reflection_table.from_file(
                dials_data("multi_crystal_proteinase_k", pathlib=True)
                / f"reflections_{i}.pickle"
            )
        )
        expts.extend(
            load.experiment_list(
                dials_data("multi_crystal_proteinase_k", pathlib=True)
                / f"experiments_{i}.json",
                check_format=False,
            )
        )
    # first make sure ids are set up correctly.
    experiments, reflections = assign_unique_identifiers(expts, refls)
    reflections = update_imageset_ids(experiments, reflections)
    joint_reflections = flex.reflection_table()
    for refls in reflections:
        joint_reflections.extend(refls)
    # check that there are 8 unique id and imageset_ids, and that these
    # correctly correspond to each experiment
    assert len(set(joint_reflections["id"])) == 8
    assert len(set(joint_reflections["imageset_id"])) == 8
    for id_ in range(8):
        sel = joint_reflections["id"] == id_
        assert set(joint_reflections["imageset_id"].select(sel)) == {id_}
